#!/usr/bin/env python
# Python wrapper generation
# Created by David P. Grote, March 6, 1998
# Modified by T. B. Yang, May 21, 1998

import sys
import os.path
from interfaceparser import processfile
import string
import re
import fvars
import pickle
from Forthon_options import args
from cfinterface import *
import wrappergen_derivedtypes
if sys.hexversion >= 0x20501f0:
    import hashlib
else:
    # --- hashlib was not available in python earlier than 2.5.
    import md5 as hashlib


class PyWrap:
    """
    Usage:
      -a       All groups will be allocated on initialization
      -t ARCH  Build for specified architecture (default is HP700)
      -d <.scalars file>  a .scalars file in another module that this module depends on
      -F <compiler> The fortran compiler being used. This is needed since some
                    operations depend on the compiler specific oddities.
      --nowritemodules The modules will not be written out, assuming
                       that they are already written.
      --macros pkg.v Other interface files that are needed for the definition
                     of macros.
      --timeroutines Calls to the routines from python will be timed
      file1    Main variable description file for the package
      [file2, ...] Subsidiary variable description files
    """

    def __init__(self, varfile, pkgname, pkgsuffix, pkgbase, initialgallot=1, writemodules=1,
                 otherinterfacefiles=[], other_scalar_vars=[], timeroutines=0,
                 otherfortranfiles=[], fcompname=None):
        self.varfile = varfile
        self.pkgname = pkgname
        self.pkgsuffix = pkgsuffix
        self.pkgbase = pkgbase
        self.initialgallot = initialgallot
        self.writemodules = writemodules
        self.timeroutines = timeroutines
        self.otherinterfacefiles = otherinterfacefiles
        self.other_scalar_vars = other_scalar_vars
        self.otherfortranfiles = otherfortranfiles
        self.fcompname = fcompname
        self.isz = isz  # isz defined in cfinterface

        self.processvariabledescriptionfile()

    def cname(self, n):
        # --- Standard name of the C interface to a Fortran routine
        # --- pkg_varname
        return self.pkgname + '_' + n

    transtable = (10*string.ascii_lowercase)[:256]

    def fsub(self, prefix, suffix=''):
        """
        The fortran standard limits routine names to 31 characters. If the
        routine name is longer than that, this routine takes the first 15
        characters and and creates a hashed string based on the full name to get
        the next 16. This does not guarantee uniqueness, but the nonuniqueness
        should be minute.
        """
        name = self.pkgname + prefix + suffix
        if len(name) < 32:
            return name
        transtable = PyWrap.transtable
        if sys.hexversion >= 0x03000000:
            hashbytes = hashlib.md5(name.encode()).digest()
            hash = ''.join([transtable[d] for d in hashbytes])
        else:
            hash = hashlib.md5(name).digest().translate(transtable)
        return name[:15] + hash

    def dimisparameter(self, dim):
        # --- Convert fortran variable name into reference from list of variables
        # --- and check if it is a parameter.
        sl = re.split('[ ()/\*\+\-]', dim)
        for ss in sl:
            if re.search('[a-zA-Z]', ss) is not None:
                try:
                    v = self.slist[self.sdict[ss]]
                    if v.parameter:
                        return True
                except KeyError:
                    pass
        return False

    def prefixdimsc(self, dim):
        # --- Convert fortran variable name into reference from list of variables.
        sl = re.split('[ ()/\*\+\-]', dim)
        for ss in sl:
            if re.search('[a-zA-Z]', ss) is not None:
                if ss in self.sdict:
                    dim = re.sub(ss,
                                 '*(long *)' + self.pkgname + '_fscalars[' + repr(self.sdict[ss]) + '].data',
                                 dim, count=1)
                else:
                    for other_vars in self.other_scalar_vars:
                        other_dict = other_vars[0]
                        if ss in other_dict:
                            dim = re.sub(ss, '*(long *)' + other_dict['_module_name_'] +
                                         '_fscalars[' + repr(other_dict[ss]) + '].data', dim, count=1)
                            break
                    else:
                        raise SyntaxError(ss + ' is not declared in the interface file')
        return dim.lower()

    # --- Convert dimensions for unspecified arrays
    def prefixdimsf(self, dim):
        # --- Check for any unspecified dimensions and replace it with an element
        # --- from the dims array.
        sl = re.split(',', dim[1:-1])
        for i in range(len(sl)):
            if sl[i] == ':':
                sl[i] = 'dims__(%d)'%(i + 1)
        dim = '(' + ','.join(sl) + ')'
        return dim.lower()

    def dimsgroups(self, dim):
        # --- Returns a list of group names that contain the variables listed in
        # --- a dimension statement
        groups = []
        sl = re.split('[ (),:/\*\+\-]', dim)
        for ss in sl:
            if re.search('[a-zA-Z]', ss) is not None:
                if ss in self.sdict:
                    groups.append(self.slist[self.sdict[ss]].group)
                else:
                    for other_vars in self.other_scalar_vars:
                        other_dict = other_vars[0]
                        other_list = other_vars[1]
                        if ss in other_dict:
                            groups.append(other_list[other_dict[ss]].group)
                            break
                    else:
                        raise SyntaxError(ss + ' is not declared in the interface file')
        return groups

    def getmodulename(self):
        if self.pkgbase is not None:
            return self.pkgbase
        else:
            return self.pkgname + self.pkgsuffix + 'py'

    def cw(self, text, noreturn=0):
        if noreturn:
            self.cfile.write(text)
        else:
            self.cfile.write(text + '\n')

    def fw(self, text, noreturn=0):
        i = 0
        while len(text[i:]) > 132 and text[i:].find('&') == -1:
            # --- If the line is too long, then break it up, adding line
            # --- continuation marks in between any variable names.
            # --- This is the same as \W, but also skips %, since PG compilers
            # --- don't seem to like a line continuation mark just before a %.
            ss = re.search('[^a-zA-Z0-9_%]', text[i+130::-1])
            assert ss is not None, "Forthon can't find a place to break up this line:\n" + text
            text = text[:i+130-ss.start()] + '&\n' + text[i+130-ss.start():]
            i += 130 - ss.start() + 1
        if noreturn:
            self.ffile.write(text)
        else:
            self.ffile.write(text + '\n')

    def setffile(self):
        """
        Set the ffile attribute, which is the fortran file object.
        It the attribute hasn't been created, then open the file with write status.
        If it has, and the file is closed, then open it with append status.
        """
        if 'ffile' in self.__dict__:
            status = 'a'
        else:
            status = 'w'
        if status == 'w' or (status == 'a' and self.ffile.closed):
            self.ffile = open(self.pkgname + '_p.F90', status)

    def processvariabledescriptionfile(self):
        """
        Read in and parse the variable description file and create the lists
        of scalars and arrays.
        """

        # --- Get the list of variables and subroutine from the var file
        vlist, hidden_vlist, typelist = processfile(self.pkgname, self.varfile,
                                                    self.otherinterfacefiles,
                                                    self.timeroutines)

        # --- Get a list of all of the group names which have variables in it
        # --- (only used when writing fortran files but done here while complete
        # --- list of variables is still in one place, vlist).
        currentgroup = ''
        groups = []
        for v in vlist:
            if not v.function and v.group != currentgroup:
                groups.append(v.group)
                currentgroup = v.group

        # --- Get a list of all of the hidden group names.
        current_hidden_group = ''
        hidden_groups = []
        for hv in hidden_vlist:
            if not hv.function and hv.group != current_hidden_group:
                hidden_groups.append(hv.group)
                current_hidden_group = hv.group

        # --- Select out all of the scalars and build a dictionary
        # --- The dictionary is used to get number of the variables use as
        # --- dimensions for arrays.
        slist = []
        sdict = {}
        i = 0
        temp = vlist[:]
        for v in temp:
            if not v.dims and not v.function:
                slist.append(v)
                sdict[v.name] = i
                i = i + 1
                vlist.remove(v)

        # --- Select out all of the arrays
        alist = []
        i = 0
        temp = vlist[:]
        for v in temp:
            if not v.function:
                alist.append(v)
                i = i + 1
                vlist.remove(v)
        temp = []

        # --- The remaining elements should all be functions
        flist = vlist

        self.typelist = typelist
        self.slist = slist
        self.sdict = sdict
        self.alist = alist
        self.flist = flist
        self.groups = groups
        self.hidden_groups = hidden_groups

    ###########################################################################
    def createmodulefile(self):
        # --- This is the routine that does all of the work

        # --- Create the module file
        self.cfile = open(self.pkgname + 'pymodule.c', 'w')
        self.cw('#include "Forthon.h"')
        self.cw('#include <setjmp.h>')
        self.cw('ForthonObject *' + self.pkgname + 'Object;')

        # --- See the kaboom command in Forthon.c for information on these two
        # --- variables.
        self.cw('extern jmp_buf stackenvironment;')
        self.cw('extern int lstackenvironmentset;')

        # --- Print out the external commands
        self.cw('extern void ' + fname(self.fsub('passpointers')) + '(void);')
        self.cw('extern void ' + fname(self.fsub('nullifypointers')) + '(void);')

        # --- fortran routine prototypes
        for f in self.flist:
            # --- Functions
            self.cw('extern ' + fvars.ftoc(f.type) + ' ' + fname(f.name) + '(', noreturn=1)
            i = 0
            istr = 0
            if len(f.args) == 0:
                self.cw('void', noreturn=1)
            for a in f.args:
                if i > 0:
                    self.cw(', ', noreturn=1)
                i = i + 1
                self.cw(fvars.ftoc(a.type) + ' ', noreturn=1)
                if a.type == 'string' or a.type == 'character':
                    istr = istr + 1
                else:
                    self.cw('*', noreturn=1)
                self.cw(a.name, noreturn=1)
            if charlen_at_end:
                for i in range(istr):
                    self.cw(', int sl' + repr(i), noreturn=1)
            self.cw(');')
        for t in self.typelist:
            self.cw('extern PyObject *' + self.cname(t.name) + 'New(PyObject *self, PyObject *args);')
        self.cw('')

        # --- setpointer and getpointer routines
        # --- Note that setpointer gets written out for all derived types -
        # --- for non-dynamic derived types, the setpointer routine does a copy.
        # --- The double underscores in the argument names are to avoid name
        # --- collisions with package variables.
        for s in self.slist:
            if (s.dynamic or s.derivedtype) and not s.parameter:
                self.cw('extern void ' + fname(self.fsub('setscalarpointer', s.name)) +
                        '(char *p__, char *fobj__, npy_intp *nullit__);')
            if s.dynamic:
                self.cw('extern void ' + fname(self.fsub('getscalarpointer', s.name)) +
                        '(ForthonObject **cobj__, char *fobj__, int *createnew__);')
        for a in self.alist:
            self.cw('extern void ' + fname(self.fsub('setarraypointer', a.name)) +
                    '(char *p__, char *fobj__, npy_intp *dims__);')
            if re.search('fassign', a.attr):
                self.cw('extern void ' + fname(self.fsub('getarraypointer', a.name)) +
                        '(Fortranarray *farray__, char *fobj__);')
        self.cw('')

        # --- setaction and getaction routines
        for s in self.slist:
            if s.setaction is not None:
                self.cw('extern void ' + fname(self.fsub('setaction', s.name)) +
                        '(' + fvars.ftoc_dict[s.type] + ' *v);')
            if s.getaction is not None:
                self.cw('extern void ' + fname(self.fsub('getaction', s.name)) + '(void);')
        for a in self.alist:
            if a.setaction is not None:
                self.cw('extern void ' + fname(self.fsub('setaction', a.name)) +
                        '(' + fvars.ftoc_dict[a.type] + ' *v);')
            if a.getaction is not None:
                self.cw('extern void ' + fname(self.fsub('getaction', a.name)) + '(void);')

        ###########################################################################
        # --- Write declarations of c pointers to fortran variables

        # --- Declare scalars from other modules
        for other_vars in self.other_scalar_vars:
            other_dict = other_vars[0]
            self.cw('extern Fortranscalar ' + other_dict['_module_name_'] +
                    '_fscalars[];')

        # --- Note that the pointers to the subroutines set and getpointer and
        # --- set and get action are at first set to NULL. The data is then setup
        # --- in a seperate call to the declarevars. This is done since it is not
        # --- standard C to have function pointers in structure initializers.

        # --- Scalars
        self.cw('int ' + self.pkgname + 'nscalars = ' + repr(len(self.slist)) + ';')
        if len(self.slist) > 0:
            self.cw('Fortranscalar ' + self.pkgname + '_fscalars[' + repr(len(self.slist)) + ']={')
            for i in range(len(self.slist)):
                s = self.slist[i]
                self.cw('{NPY_%s,'%fvars.ftop(s.type) +
                        '"%s",'%s.type +
                        '"%s",'%s.name +
                        'NULL,' +
                        '"%s",'%s.group +
                        '"%s",'%s.attr +
                        '"%s",'%repr(s.comment)[1:-1].replace('"', '\\"') +
                        '%i,'%s.dynamic +
                        '%i,'%s.parameter +
                        'NULL,' +  # setscalarpointer
                        'NULL,' +  # getscalarpointer
                        'NULL,' +  # setaction
                        'NULL' +  # getaction
                        '}', noreturn=1)
                if i < len(self.slist) - 1:
                    self.cw(',')
            self.cw('};')
        else:
            self.cw('Fortranscalar *' + self.pkgname + '_fscalars=NULL;')

        # --- Arrays
        self.cw('int ' + self.pkgname + 'narrays = ' + repr(len(self.alist)) + ';')
        if len(self.alist) > 0:
            self.cw('static Fortranarray ' +
                    self.pkgname + '_farrays[' + repr(len(self.alist)) + ']={')
            for i in range(len(self.alist)):
                a = self.alist[i]
                if a.data and a.dynamic:
                    initvalue = a.data[1:-1]
                else:
                    initvalue = '0'
                self.cw('{NPY_%s,'%fvars.ftop(a.type) +
                        '%d,'%a.dynamic +
                        '%d,'%len(a.dims) +
                        'NULL,' +
                        '"%s",'%a.name +
                        '{NULL},' +
                        'NULL,' +  # setarraypointer
                        'NULL,' +  # getarraypointer
                        'NULL,' +  # setaction
                        'NULL,' +  # getaction
                        '%s,'%initvalue +
                        'NULL,' +
                        '"%s",'%a.group +
                        '"%s",'%a.attr +
                        '"%s",'%repr(a.comment)[1:-1].replace('"', '\\"') +
                        '"%s"}'%repr(a.dimstring)[1:-1], noreturn=1)
                if i < len(self.alist) - 1:
                    self.cw(',')
            self.cw('};')
        else:
            self.cw('static Fortranarray *' + self.pkgname + '_farrays=NULL;')

        #########################################################################
        # --- Write declarations of c pointers to fortran variables
        self.cw('void ' + self.pkgname + 'declarevars(ForthonObject *obj) {')

        # --- Scalars
        for i in range(len(self.slist)):
            s = self.slist[i]
            if (s.dynamic or s.derivedtype) and not s.parameter:
                setscalarpointer = '*' + fname(self.fsub('setscalarpointer', s.name))
                self.cw('obj->fscalars[%d].setscalarpointer = %s;'%(i, setscalarpointer))
                if s.dynamic:
                    getscalarpointer = '*' + fname(self.fsub('getscalarpointer', s.name))
                    self.cw('obj->fscalars[%d].getscalarpointer = %s;'%(i, getscalarpointer))
            if s.setaction is not None:
                setaction = '*' + fname(self.fsub('setaction', s.name))
                self.cw('obj->fscalars[%d].setaction = %s;'%(i, setaction))
            if s.getaction is not None:
                getaction = '*' + fname(self.fsub('getaction', s.name))
                self.cw('obj->fscalars[%d].getaction = %s;'%(i, getaction))

        # --- Arrays
        for i in range(len(self.alist)):
            a = self.alist[i]
            if a.dynamic:
                setarraypointer = '*' + fname(self.fsub('setarraypointer', a.name))
                self.cw('obj->farrays[%d].setarraypointer = %s;'%(i, setarraypointer))
            if re.search('fassign', a.attr):
                getarraypointer = '*' + fname(self.fsub('getarraypointer', a.name))
                self.cw('obj->farrays[%d].getarraypointer = %s;'%(i, getarraypointer))
            if a.setaction is not None:
                setaction = '*' + fname(self.fsub('setaction', a.name))
                self.cw('obj->farrays[%d].setaction = %s;'%(i, setaction))
            if a.getaction is not None:
                getaction = '*' + fname(self.fsub('getaction', a.name))
                self.cw('obj->farrays[%d].getaction = %s;'%(i, getaction))

        self.cw('}')

# Some extra work is needed to get the getset attribute access scheme working.
#   # --- Write out the table of getset routines
#   self.cw('')
#   self.cw('static PyGetSetDef ' + self.pkgname + '_getseters[] = {')
#   for i in range(len(self.slist)):
#     s = self.slist[i]
#     if s.type == 'real': gstype = 'double'
#     elif s.type == 'double': gstype = 'double'
#     elif s.type == 'float': gstype = 'float'
#     elif s.type == 'integer': gstype = 'integer'
#     elif s.type == 'complex': gstype = 'cdouble'
#     else:                    gstype = 'derivedtype'
#     self.cw('{"' + s.name + '", (getter)Forthon_getscalar' + gstype +
#                          ', (setter)Forthon_setscalar' + gstype +
#                    ', "%s"'%repr(s.comment)[1:-1].replace('"', '\\"') +
#                         ', (void *)' + repr(i) + '}, ')
#   for i in range(len(self.alist)):
#     a = self.alist[i]
#     self.cw('{"' + a.name + '", (getter)Forthon_getarray' +
#                          ', (setter)Forthon_setarray' +
#                    ', "%s"'%repr(a.comment)[1:-1].replace('"', '\\"') +
#                         ', (void *)' + repr(i) + '}, ')
#   self.cw('{"scalardict", (getter)Forthon_getscalardict, ' +
#                         '(setter)Forthon_setscalardict, ' +
#           '"internal scalar dictionary", NULL}, ')
#   self.cw('{"arraydict", (getter)Forthon_getarraydict, ' +
#                        '(setter)Forthon_setarraydict, ' +
#           '"internal array dictionary", NULL}, ')
#   self.cw('{NULL}};')

        ###########################################################################
        if self.timeroutines:
            # --- This is written out here instead of just being in Forthon.h
            # --- so that when it is not used, the compiler doesn't complain
            # --- about cputime being unused.
            self.cw('#include <sys/times.h>')
            self.cw('#include <unistd.h>')
            self.cw('static double cputime(void)')
            self.cw('{')
            self.cw('  struct tms usage;')
            self.cw('  long hardware_ticks_per_second;')
            self.cw('  (void) times(&usage);')
            self.cw('  hardware_ticks_per_second = sysconf(_SC_CLK_TCK);')
            self.cw('  return (double) usage.tms_utime/hardware_ticks_per_second;')
            self.cw('}')

        ###########################################################################
        ###########################################################################
        # --- Now, the fun part, writing out the wrapper for the subroutine and
        # --- function calls.
        for f in self.flist:
            # --- Write out the documentation first.
            docstring = ('static char doc_' + self.cname(f.name) + '[] = "' + f.name +
                         f.dimstring + '\n' + f.comment + '";')
            # --- Replaces newlines with '\\n' so that the string is all on one line
            # --- in the C coding using repr, easily.
            # --- The [1:-1] strips off the single quotes that repr puts there.
            self.cw(repr(docstring)[1:-1])
            # --- Now write out the wrapper
            self.cw('static PyObject *')
            self.cw(self.cname(f.name) + '(PyObject *self, PyObject *args)')
            self.cw('{')

            # --- With arguments, it gets very messy
            lv = repr(len(f.args))
            if len(f.args) > 0:
                self.cw('  PyObject * pyobj[' + lv + '];')
                self.cw('  PyArrayObject * ax[' + lv + '];')
                self.cw('  int i;')
                self.cw('  char e[256];')

            if self.timeroutines:
                # --- Setup for the timer, getting time routine started.
                self.cw('  double time1, time2;')
                self.cw('  time1 = cputime();')

            # --- For character arguments, need to create an FSTRING array.
            istr = 0
            for a in f.args:
                if a.type == 'string' or a.type == 'character':
                    istr = istr + 1
            if istr > 0:
                self.cw('  FSTRING fstr[' + repr(istr) + '];')

            # --- If this is a function, set up variables to hold return value
            if f.type != 'void':
                self.cw('  PyObject * ret_val;')
                self.cw('  ' + fvars.ftoc(f.type) + ' r;')

            # --- Set all of the ax's to NULL
            if len(f.args) > 0:
                self.cw('  for (i=0;i<' + repr(len(f.args)) + ';i++) ax[i] = NULL;')

            # --- Parse incoming arguments into a list of PyObjects
            self.cw('  if (!PyArg_ParseTuple(args, "' + 'O'*len(f.args) + '"', noreturn=1)
            for i in range(len(f.args)):
                self.cw(', &pyobj[' + repr(i) + ']', noreturn=1)
            self.cw(')) return NULL;')

            # --- Loop over arguments, extracting the data addresses.
            # --- Convert all arguments into arrays. This allows complete flexibility
            # --- in what can be passed to fortran functions.
            istr = 0
            for i in range(len(f.args)):
                if not fvars.isderivedtype(f.args[i]):
                    self.cw('  if (!Forthon_checksubroutineargtype(pyobj[' + repr(i) + '], ' +
                            'NPY_' + fvars.ftop(f.args[i].type) + ')) {')
                    self.cw('    sprintf(e, "Argument ' + f.args[i].name + ' in ' + f.name +
                            ' has the wrong type");')
                    self.cw('    PyErr_SetString(ErrorObject, e);')
                    self.cw('    goto err;}')
                    if f.function == 'fsub':
                        self.cw('  ax[' + repr(i) + '] = FARRAY_FROMOBJECT(' +
                                'pyobj[' + repr(i) + '], NPY_' + fvars.ftop(f.args[i].type) + ');')
                    elif f.function == 'csub':
                        self.cw('  ax[' + repr(i) + ']=(PyArrayObject *)PyArray_ContiguousFromObject(' +
                                'pyobj[' + repr(i) + '], NPY_' + fvars.ftop(f.args[i].type) + ', 0, 0);')
                    self.cw('  if (ax[' + repr(i) + '] == NULL) {')
                    self.cw('    sprintf(e, "There is an error in argument ' + f.args[i].name +
                            ' in ' + f.name + '");')
                    self.cw('    PyErr_SetString(ErrorObject, e);')
                    self.cw('    goto err;}')
                    if f.args[i].type == 'string' or f.args[i].type == 'character':
                        self.cw(' FSETSTRING(fstr[%d], PyArray_BYTES(ax[%d]), PyArray_ITEMSIZE(ax[%d]));'%(istr, i, i))
                        istr = istr + 1
                else:
                    self.cw('  {')
                    self.cw('  PyObject *t;')
                    self.cw('  t = PyObject_Type(pyobj[' + repr(i) + ']);')
                    self.cw('  if (strcmp(((PyTypeObject *)t)->tp_name, "Forthon") != 0) {')
                    self.cw('    sprintf(e, "Argument ' + f.args[i].name + ' in ' + f.name +
                            ' has the wrong type");')
                    self.cw('    PyErr_SetString(ErrorObject, e);')
                    self.cw('    goto err;}')
                    self.cw('  Py_DECREF(t);')
                    typename = '((ForthonObject *)pyobj[' + repr(i) + '])->typename'
                    self.cw('  if (strcmp(' + typename + ', "' + f.args[i].type + '") != 0) {')
                    self.cw('    sprintf(e, "Argument ' + f.args[i].name + ' in ' + f.name +
                            ' has the wrong type");')
                    self.cw('    PyErr_SetString(ErrorObject, e);')
                    self.cw('    goto err;}')
                    self.cw('  }')

            # --- Write the code checking dimensions of arrays
            # --- This must be done after all of the ax's are setup in case the
            # --- dimensioning arguments come after the array argument.
            # --- This creates a local variable with the same name as the argument
            # --- and gives it the value passed in. Then any expressions in the
            # --- dimensions statements will be explicitly evaluated in C.
            if len(f.dimvars) > 0:
                self.cw('  {')
                self.cw('  long _n;')
                # --- Declare the dimension variables.
                for var, i in f.dimvars:
                    self.cw('  ' + fvars.ftoc(var.type) + ' ' + var.name + '=*' +
                            '(' + fvars.ftoc(var.type) + ' *)(PyArray_BYTES(ax[' + repr(i) + ']));')
                # --- Loop over the arguments, looking for dimensioned arrays
                i = -1
                for arg in f.args:
                    i += 1
                    if len(arg.dims) > 0:
                        # --- Check the rank of the input argument
                        # --- For a 1-D argument, allow a scaler to be passed, which has
                        # --- a number of dimensions (nd) == 0.
                        self.cw('  if (!(PyArray_NDIM(ax[%d]) == %d'%(i, len(arg.dims)), noreturn=1)
                        if len(arg.dims) == 1:
                            self.cw('      ||PyArray_NDIM(ax[%d]) == 0)) {'%i)
                        else:
                            self.cw('      )) {')
                        self.cw('    sprintf(e, "Argument %s in %s '%(arg.name, f.name) +
                                'has the wrong number of dimensions");')
                        self.cw('    PyErr_SetString(ErrorObject, e);')
                        self.cw('    goto err;}')

                        # --- Skip the check of dimension sizes if the total size of
                        # --- the array should be zero. This gets around an issue with
                        # --- numpy that zero length arrays seem to be always in
                        # --- C ordering.
                        self.cw('  if (1', noreturn=1)
                        for dim in arg.dims:
                            self.cw('*((' + dim.high + ') - (' + dim.low + ') + 1)', noreturn=1)
                        self.cw(' != 0) {')

                        j = -1
                        for dim in arg.dims:
                            j += 1
                            # --- Compare each dimension with its specified value
                            # --- For a 1-D argument, allow a scalar to be passed, which has
                            # --- a number of dimensions (nd) == 0, but only if the
                            # --- argument needs to have a length of 0 or 1.
                            self.cw('    _n = (' + dim.high + ') - (' + dim.low + ') + 1;')
                            if len(arg.dims) == 1:
                                self.cw('    if (!((_n==0||_n==1)||(PyArray_NDIM(ax[%d]) > 0 &&'%i,
                                        noreturn=1)
                            else:
                                self.cw('    if (!((', noreturn=1)
                            self.cw('_n == (long)(PyArray_DIMS(ax[%d])[%d])))) {'%(i, j))
                            self.cw('      sprintf(e, "Dimension ' + repr(j + 1) + ' of argument ' +
                                    arg.name + ' in ' + f.name +
                                    ' has the wrong size");')
                            self.cw('      PyErr_SetString(ErrorObject, e);')
                            self.cw('      goto err;}')
                        self.cw('  }')
                self.cw('  }')

            # --- If the stackenvironment has not already been set, then make a call
            # --- to setjmp to save the state in case an error happens.
            # --- If there was an error, setjmp returns 1, so exit out.
            # --- If this routine is called from a python routine that was called
            # --- from another fortran routine, then the stackenvironment will
            # --- already have been setup (from the calling fortran routine) so don't
            # --- reset it.
            self.cw('  if (!(lstackenvironmentset++) && setjmp(stackenvironment)) goto err;')

            # --- Write the actual call to the fortran routine.
            if f.type == 'void':
                self.cw('  ')
            else:
                self.cw('  r = ')
            self.cw(fname(f.name) + '(', noreturn=1)
            i = 0
            istr = 0
            for a in f.args:
                if i > 0:
                    self.cw(', ', noreturn=1)
                if fvars.isderivedtype(a):
                    self.cw('((ForthonObject *)(pyobj[' + repr(i) + ']))->fobj', noreturn=1)
                elif a.type == 'string' or a.type == 'character':
                    self.cw('fstr[%d]'%(istr), noreturn=1)
                    istr = istr + 1
                else:
                    self.cw('(' + fvars.ftoc(a.type) + ' *)(PyArray_BYTES(ax[' + repr(i) + ']))', noreturn=1)
                i = i + 1
            if charlen_at_end:
                i = 0
                istr = 0
                for a in f.args:
                    if a.type == 'string' or a.type == 'character':
                        self.cw(', (int)PyArray_ITEMSIZE(ax[' + repr(i) + '])', noreturn=1)
                        istr = istr + 1
                    i = i + 1

            self.cw(');')  # --- Closing parenthesis on the call list

            # --- Decrement the counter. This will reach zero when the top of the
            # --- fortran call chain is reached and is about to return to the top
            # --- level python.
            self.cw('  lstackenvironmentset--;')

            # --- Copy the data that was sent to the routine back into the passed
            # --- in object if it is an PyArray.
            # --- Decrement reference counts of array objects created.
            # --- This is now handled by a separate subroutine included in Forthon.h
            if len(f.args) > 0:
                self.cw('  Forthon_restoresubroutineargs(' + repr(len(f.args)) + ', pyobj, ax);')

            if self.timeroutines:
                # --- Now get ending time and add to timer variable
                self.cw('  time2 = cputime();')
                self.cw('  *(double *)' + self.pkgname + '_fscalars[' +
                        repr(self.sdict[f.name + 'runtime']) + '].data += (time2 - time1);')

            # --- Write return sequence
            if f.type == 'void':
                self.cw('  returnnone;')
            else:
                self.cw('  ret_val = Py_BuildValue ("' + fvars.fto1[f.type] + '", r);')
                self.cw('  return ret_val;')

            # --- Error section, in case there was an error above or in the
            # --- fortran call
            self.cw('err:')

            if len(f.args) > 0:
                # --- Decrement reference counts of array objects created.
                self.cw('  for (i=0;i<' + repr(len(f.args)) + ';i++)')
                self.cw('    if (ax[i] != NULL) {Py_XDECREF(ax[i]);}')

            self.cw('  return NULL;')

            self.cw('}')

        # --- Add blank line
        self.cw('')

        ###########################################################################
        # --- Write out method list
        self.cw('static struct PyMethodDef ' + self.pkgname + '_methods[] = {')
        for f in self.flist:
            if f.function:
                self.cw('{"' + f.name + '", (PyCFunction)' + self.cname(f.name) + ', 1, ' +
                        'doc_' + self.cname(f.name) + '}, ')
        for t in self.typelist:
            self.cw('{"' + t.name + '", (PyCFunction)' + self.cname(t.name) + 'New, 1, ' +
                    '"Creates a new instance of fortran derived type ' + t.name + '"}, ')
        self.cw('{NULL, NULL}};')
        self.cw('')

        ###########################################################################
        # --- Write static array initialization routines
        self.cw('void ' + self.pkgname + 'setstaticdims(ForthonObject *self)')
        self.cw('{')

        i = -1
        for a in self.alist:
            i = i + 1
            vname = self.pkgname + '_farrays[' + repr(i) + ']'
            if a.dims and not a.dynamic:
                j = 0
                for d in a.dims:
                    if d.high == '':
                        continue
                    self.cw('   ' + vname + '.dimensions[' + repr(j) + ']=(npy_intp)((int)',
                            noreturn=1)
                    j = j + 1
                    if re.search('[a-zA-Z]', d.high) is None:
                        self.cw('(' + d.high + ') - ', noreturn=1)
                    else:
                        if not self.dimisparameter(d.high):
                            raise SyntaxError('%s: static dims must be constants or parameters'%a.name)
                        self.cw('(' + self.prefixdimsc(d.high) + ') - ', noreturn=1)
                    if re.search('[a-zA-Z]', d.low) is None:
                        self.cw('(' + d.low + ') + 1);')
                    else:
                        if not self.dimisparameter(d.low):
                            raise SyntaxError('%s: static dims must be constants or parameters'%a.name)
                        self.cw('(' + self.prefixdimsc(d.low) + ') + 1);')

        self.cw('}')
        self.cw('')

        ###########################################################################
        # --- Write routine which sets the dimensions of the dynamic arrays.
        # --- This is done in a seperate routine so it only appears once.
        # --- A routine is written out for each group which has dynamic arrays. Then
        # --- a routine is written which calls all of the individual group routines.
        # --- That is done to reduce the strain on the compiler by reducing the size
        # --- of the routines. (In fact, in one case, with everything in one
        # --- routine the cc compiler was giving a core dump!)
        # --- Loop over the variables. This assumes that the variables are sorted
        # --- by group.
        i = -1
        currentgroup = ''
        dyngroups = []
        for a in self.alist:
            if a.group != currentgroup and a.dynamic:
                if currentgroup != '':
                    self.cw('  }}')
                currentgroup = a.group
                if len(dyngroups) > 0:
                    dyngroups[-1][2] = i
                dyngroups.append([currentgroup, i + 1, len(self.alist)])
                self.cw('static void ' + self.pkgname + 'setdims' + currentgroup + '(char *name, long i)')
                self.cw('{')
                self.cw('  if (strcmp(name, "' + a.group + '") || strcmp(name, "*")) {')

            i = i + 1
            vname = self.pkgname + '_farrays[' + repr(i) + ']'
            if a.dynamic == 1 or a.dynamic == 2:
                j = 0
                self.cw('  if (i == -1 || i == %d) {'%i)
                # --- create lines of the form dims[1] = high - low + 1
                for d in a.dims:
                    if d.high == '':
                        continue
                    self.cw('   ' + vname + '.dimensions[' + repr(j) + ']=(npy_intp)((int)',
                            noreturn=1)
                    j = j + 1
                    if re.search('[a-zA-Z]', d.high) is None:
                        self.cw('(' + d.high + ') - ', noreturn=1)
                    else:
                        self.cw('(' + self.prefixdimsc(d.high) + ') - ', noreturn=1)
                    if re.search('[a-zA-Z]', d.low) is None:
                        self.cw('(' + d.low + ') + 1);')
                    else:
                        self.cw('(' + self.prefixdimsc(d.low) + ') + 1);')
                self.cw('  }')

        if currentgroup != '':
            self.cw('  }}')

        # --- Now write out the setdims routine which calls of the routines
        # --- for the individual groups.
        self.cw('void ' + self.pkgname + 'setdims(char *name, ForthonObject *obj, long i)')
        self.cw('{')
        for groupinfo in dyngroups:
            self.cw('  if (i == -1 || (%d <= i && i <= %d))'%tuple(groupinfo[1:]), noreturn=1)
            self.cw('  ' + self.pkgname + 'setdims' + groupinfo[0] + '(name, i);')
        self.cw('}')

        self.cw('')

        ###########################################################################
        # --- Write set pointers routine which gets all of the fortran pointers
        self.cw('void ' + fname(self.fsub('grabscalarpointers')) + '(long *i, char *p)')
        self.cw('{')
        self.cw('  /* Gabs pointer for the scalar */')
        self.cw('  ' + self.pkgname + '_fscalars[*i].data = (char *)p;')
        self.cw('}')

        # --- A serarate routine is needed for derived types since the cobj__
        # --- that is passed in is already a pointer, so **p is needed.
        self.cw('void ' + fname(self.fsub('setderivedtypepointers')) + '(long *i, char **p)')
        self.cw('{')
        self.cw('  /* Gabs pointer for the scalar */')
        self.cw('  ' + self.pkgname + '_fscalars[*i].data = (char *)(*p);')
        self.cw('}')

        # --- Get pointer to an array. This takes an integer to specify which array
        self.cw('void ' + fname(self.fsub('grabarraypointers')) + '(long *i, char *p)')
        self.cw('{')
        self.cw('  /* Grabs pointer for the array */')
        self.cw('  ' + self.pkgname + '_farrays[*i].data.s = (char *)p;')
        self.cw('}')

        # --- This takes a Fortranarray object directly.
        self.cw('void ' + fname(self.fsub('grabarraypointersobj')) + '(Fortranarray *farray, char *p)')
        self.cw('{')
        self.cw('  /* Grabs pointer for the array */')
        self.cw('  farray->data.s = (char *)p;')
        self.cw('}')

        # --- This routine gets the dimensions from an array. It is called from
        # --- fortran and the last argument should be shape(array).
        # --- This is only used for routines with the fassign attribute.
        # --- Note that the dimensions are stored in C order.
        self.cw('void ' + fname(self.fsub('setarraydims')) + '(Fortranarray *farray, long *dims)')
        self.cw('{')
        self.cw('  int id;')
        self.cw('  for (id=0;id<farray->nd;id++)')
        self.cw('    farray->dimensions[id] = (npy_intp)(dims[id]);')
        self.cw('}')

        ###########################################################################
        # --- And finally, the initialization function
        if sys.hexversion >= 0x03000000:
            self.cw('static struct PyModuleDef moduledef = {')
            self.cw('  PyModuleDef_HEAD_INIT,')
            self.cw('  "{0}py", /* m_name */'.format(self.pkgname + self.pkgsuffix))
            self.cw('  "{0}", /* m_doc */'.format(self.pkgname))
            self.cw('  -1,                  /* m_size */')
            self.cw('  {0}_methods,    /* m_methods */'.format(self.pkgname))
            self.cw('  NULL,                /* m_reload */')
            self.cw('  NULL,                /* m_traverse */')
            self.cw('  NULL,                /* m_clear */')
            self.cw('  NULL,                /* m_free */')
            self.cw('  };')

        self.cw('PyMODINIT_FUNC')
        if sys.hexversion >= 0x03000000:
            self.cw('PyInit_' + self.pkgname + self.pkgsuffix + 'py(void)')
        else:
            self.cw('init' + self.pkgname + self.pkgsuffix + 'py(void)')
        self.cw('{')

        self.cw('  PyObject *m;')
        if self.fcompname == 'nag':
            self.cw('  int argc; char **argv;')
            self.cw('  Py_GetArgcArgv(&argc, &argv);')
            self.cw('  f90_init(argc, argv);')
        # self.cw('  ForthonType.tp_getset = ' + self.pkgname + '_getseters;')
        # self.cw('  ForthonType.tp_methods = ' + self.pkgname + '_methods;')
        self.cw('  if (PyType_Ready(&ForthonType) < 0)')
        if sys.hexversion >= 0x03000000:
            self.cw('    return NULL;')
        else:
            self.cw('    return;')

        if sys.hexversion >= 0x03000000:
            self.cw('  m = PyModule_Create(&moduledef);')
        else:
            self.cw('  m = Py_InitModule("' + self.pkgname + self.pkgsuffix + 'py", ' + self.pkgname + '_methods);')

        # self.cw('  PyModule_AddObject(m, "' + self.pkgname + 'Type", ' + '(PyObject *)&ForthonType);')
        self.cw('  ' + self.pkgname + 'Object=(ForthonObject *)' + 'PyObject_GC_New(ForthonObject, &ForthonType);')
        #       'ForthonObject_New(NULL, NULL);')
        self.cw('  ' + self.pkgname + 'Object->name = "' + self.pkgname + '";')
        self.cw('  ' + self.pkgname + 'Object->typename = "' + self.pkgname + '";')
        self.cw('  ' + self.pkgname + 'Object->nscalars = ' + self.pkgname + 'nscalars;')
        self.cw('  ' + self.pkgname + 'Object->fscalars = ' + self.pkgname + '_fscalars;')
        self.cw('  ' + self.pkgname + 'Object->narrays = ' + self.pkgname + 'narrays;')
        self.cw('  ' + self.pkgname + 'Object->farrays = ' + self.pkgname + '_farrays;')
        self.cw('  ' + self.pkgname + 'Object->setdims = *' + self.pkgname + 'setdims;')
        self.cw('  ' + self.pkgname + 'Object->setstaticdims = *' + self.pkgname + 'setstaticdims;')
        self.cw('  ' + self.pkgname + 'Object->fmethods = ' + self.pkgname + '_methods;')
        self.cw('  ' + self.pkgname + 'Object->__module__ = Py_BuildValue("s", "%s");'%self.getmodulename())
        self.cw('  ' + self.pkgname + 'Object->fobj = NULL;')
        self.cw('  ' + self.pkgname + 'Object->fobjdeallocate = NULL;')
        self.cw('  ' + self.pkgname + 'Object->nullifycobj = NULL;')
        self.cw('  ' + self.pkgname + 'Object->allocated = 0;')
        self.cw('  ' + self.pkgname + 'Object->garbagecollected = 0;')
        self.cw('  PyModule_AddObject(m, "' + self.pkgname + '", (PyObject *)' + self.pkgname + 'Object);')
        self.cw('  ErrorObject = PyErr_NewException("' + self.pkgname + self.pkgsuffix + 'py.error", NULL, NULL);')
        self.cw('  PyModule_AddObject(m, "' + self.pkgname + 'error", ErrorObject);')
        self.cw('  PyModule_AddObject(m, "fcompname", ' + 'PyUnicode_FromString("' + self.fcompname + '"));')
        if sys.hexversion >= 0x03000000:
            self.cw('  PyModule_AddObject(m, "realsize", ' + 'PyLong_FromLong((long)%s'%realsize + '));')
        else:
            self.cw('  PyModule_AddObject(m, "realsize", ' + 'PyInt_FromLong((long)%s'%realsize + '));')
        self.cw('  if (PyErr_Occurred()) {')
        self.cw('    PyErr_Print();')
        self.cw('    Py_FatalError("can not initialize module ' + self.pkgname + '");')
        self.cw('    }')
        self.cw('  import_array();')
        self.cw('  ' + self.pkgname + 'declarevars(' + self.pkgname + 'Object);')
        self.cw('  Forthon_BuildDicts(' + self.pkgname + 'Object);')
        self.cw('  ForthonPackage_allotdims(' + self.pkgname + 'Object);')
        self.cw('  ' + fname(self.fsub('passpointers')) + '();')
        self.cw('  ' + fname(self.fsub('nullifypointers')) + '();')
        self.cw('  ForthonPackage_staticarrays(' + self.pkgname + 'Object);')
        if self.initialgallot:
            self.cw('  {')
            self.cw('  PyObject *s;')
            self.cw('  s = Py_BuildValue("(s)", "*");')
            self.cw('  ForthonPackage_gallot((PyObject *)' + self.pkgname + 'Object, s);')
            self.cw('  Py_XDECREF(s);')
            self.cw('  }')

        self.cw('  {')
        self.cw('  PyObject *m, *d, *f, *r;')
        self.cw('  r = NULL;')
        self.cw('  m = PyImport_ImportModule("Forthon");')
        self.cw('  if (m != NULL) {')
        self.cw('    d = PyModule_GetDict(m);')
        self.cw('    if (d != NULL) {')
        self.cw('      f = PyDict_GetItemString(d, "registerpackage");')
        self.cw('      if (f != NULL) {')
        self.cw('        r = PyObject_CallFunction(f, "Os", (PyObject *)' + self.pkgname + 'Object, "' + self.pkgname + '");')
        self.cw('  }}}')
        self.cw('  if (NULL == r) {')
        self.cw('    if (PyErr_Occurred()) PyErr_Print();')
        self.cw('    Py_FatalError("unable to find a compatible Forthon module in which to register module ' + self.pkgname + '");')
        self.cw('  }')
        self.cw('  Py_XDECREF(m);')
        self.cw('  Py_XDECREF(r);')
        self.cw('  }')

        if machine == 'win32':
            self.cw('  /* Initialize FORTRAN on CYGWIN */')
            self.cw(' initPGfortran();')

        if sys.hexversion >= 0x03000000:
            self.cw('  return m;')

        self.cw('}')
        self.cw('')

        ###########################################################################
        # --- Close the c package module file
        self.cfile.close()

        ###########################################################################
        ###########################################################################
        ###########################################################################
        # --- Write out fortran initialization routines
        self.setffile()
        self.ffile.close()

        ###########################################################################
        ###########################################################################
        # --- Process any derived types
        wrappergen_derivedtypes.ForthonDerivedType(self.typelist, self.pkgname, self.pkgsuffix, self.pkgbase,
                                                   self.pkgname + 'pymodule.c',
                                                   self.pkgname + '_p.F90', self.isz,
                                                   self.writemodules, self.fcompname)
        ###########################################################################
        ###########################################################################

        self.setffile()

        ###########################################################################
        # --- Write out f90 modules, including any data statements
        if self.writemodules:
            self.writef90modules()

        ###########################################################################
        self.fw('subroutine ' + self.fsub('passpointers') + '()')

        # --- Write out the Use statements
        for g in self.groups + self.hidden_groups:
            self.fw('  use ' + g)

        # --- Write out calls to c routine passing down pointers to scalars
        for i in range(len(self.slist)):
            s = self.slist[i]
            if s.dynamic:
                continue
            if s.derivedtype:
                # --- This is only called for static instances, so deallocatable is
                # --- set to false (the last argument).
                self.fw('  call init' + s.type + 'py(int(' + repr(i) + ', ' + self.isz + '), ' + s.name + ', ' +
                        s.name + '%cobj__, int(1, ' + self.isz + '), int(0, ' + self.isz + '))')
                self.fw('  call ' + self.fsub('setderivedtypepointers') + '(int(' + repr(i) + ', ' + self.isz + '), ' + s.name + '%cobj__)')
            else:
                self.fw('  call ' + self.fsub('grabscalarpointers') + '(int(' + repr(i) + ', ' + self.isz + '), ' + s.name + ')')

        # --- Write out calls to c routine passing down pointers to arrays.
        for i in range(len(self.alist)):
            a = self.alist[i]
            if not a.dynamic:
                self.fw('  call ' + self.fsub('grabarraypointers') + '(int(' + repr(i) + ', ' + self.isz + '), ' + a.name + ')')

        # --- Finish the routine
        self.fw('  return')
        self.fw('end')

        ###########################################################################
        # --- Nullifies the pointers of all dynamic variables. This is needed
        # --- since in some compilers, the associated routine returns
        # --- erroneous information if the status of a pointer is undefined.
        # --- Pointers must be explicitly nullified in order to get
        # --- associated to return a false value.
        self.fw('subroutine ' + self.fsub('nullifypointers') + '()')

        # --- Write out the Use statements
        for g in self.groups + self.hidden_groups:
            self.fw('  use ' + g)

        for i in range(len(self.slist)):
            s = self.slist[i]
            if s.dynamic:
                self.fw('  nullify(' + s.name + ')')
        for i in range(len(self.alist)):
            a = self.alist[i]
            if a.dynamic:
                self.fw('  nullify(' + a.name + ')')

        self.fw('  return')
        self.fw('end')
        ###########################################################################
        # --- Write routine for each dynamic variable which gets the pointer from the
        # --- wrapper
        for s in self.slist:
            if (s.dynamic or s.derivedtype) and not s.parameter:
                self.fw('subroutine ' + self.fsub('setscalarpointer', s.name) + '(p__, fobj__, nullit__)')
                self.fw('  use ' + s.group)
                self.fw('  integer(' + self.isz + '):: fobj__')
                self.fw('  integer(' + self.isz + '):: nullit__')
                if s.type == 'character':
                    self.fw('  character(len=' + s.dims[0].high + '), target:: p__')
                else:
                    self.fw('  ' + fvars.ftof(s.type) + ', target:: p__')
                if s.dynamic:
                    self.fw('  if (nullit__ == 0) then')
                    self.fw('    ' + s.name + ' => p__')
                    self.fw('  else')
                    self.fw('    nullify(' + s.name + ')')
                    self.fw('  endif')
                else:
                    self.fw('  ' + s.name + ' = p__')
                self.fw('  return')
                self.fw('end')
            if s.dynamic:
                # --- In all cases, it is not desirable to create a new instance,
                # --- for example when the object is being deleted.
                self.fw('subroutine ' + self.fsub('getscalarpointer', s.name) +
                        '(cobj__, fobj__, createnew__)')
                self.fw('  use ' + s.group)
                self.fw('  integer(' + self.isz + '):: cobj__, fobj__')
                self.fw('  integer(4):: createnew__')
                self.fw('  if (associated(' + s.name + ')) then')
                self.fw('    if (' + s.name + '%cobj__ == 0 .and. createnew__ == 1) then')
                self.fw('      call init' + s.type + 'py(int(-1, ' + self.isz + '), ' + s.name + ', ' +
                        s.name + '%cobj__, int(0, ' + self.isz + '), int(0, ' + self.isz + '))')
                self.fw('    endif')
                self.fw('    cobj__ = ' + s.name + '%cobj__')
                self.fw('  else')
                self.fw('    cobj__ = 0')
                self.fw('  endif')
                self.fw('  return')
                self.fw('end')

        for a in self.alist:
            if a.dynamic:
                self.fw('subroutine ' + self.fsub('setarraypointer', a.name) + '(p__, fobj__, dims__)')
                groups = self.dimsgroups(a.dimstring)
                groupsprinted = [a.group]
                for g in groups:
                    if g not in groupsprinted:
                        self.fw('  use ' + g)
                        groupsprinted.append(g)
                self.fw('  use ' + a.group)
                self.fw('  integer(' + self.isz + '):: fobj__')
                self.fw('  integer(' + self.isz + '):: dims__(' + repr(len(a.dims)) + ')')

                if a.type == 'character':
                    self.fw('  character(len=' + a.dims[0].high + '), target:: p__' +
                            self.prefixdimsf(re.sub('[ \t\n]', '', a.dimstring)))
                else:
                    self.fw('  ' + fvars.ftof(a.type) + ', target:: p__' +
                            self.prefixdimsf(re.sub('[ \t\n]', '', a.dimstring)))
                self.fw('  ' + a.name + ' => p__')
                self.fw('  return')
                self.fw('end')
                if re.search('fassign', a.attr):
                    self.fw('subroutine ' + self.fsub('getarraypointer', a.name) + '(farray__, fobj__)')
                    self.fw('  use ' + a.group)
                    self.fw('  integer(' + self.isz + '):: farray__, fobj__')
                    self.fw('  integer(' + self.isz + '):: ss(%d)'%(len(a.dims)))
                    self.fw('  if (.not. associated(' + a.name + ')) return')
                    self.fw('  call ' + self.fsub('grabarraypointersobj') + '(farray__, ' + a.name + ')')
                    self.fw('  ss = shape(' + a.name + ')')
                    self.fw('  call ' + self.fsub('setarraydims') + '(farray__, ss)')
                    self.fw('  return')
                    self.fw('end')

        ###########################################################################

        # --- Close fortran file
        self.ffile.close()

        scalar_pickle_file = open(self.pkgname + '.scalars', 'wb')
        self.sdict['_module_name_'] = self.pkgname
        pickle.dump(self.sdict, scalar_pickle_file)
        pickle.dump(self.slist, scalar_pickle_file)
        scalar_pickle_file.close()

    def writef90modules(self):
        """
        Write the fortran90 modules
        """
        self.setffile()
        if self.fcompname == 'xlf':
            save = ', save'
        else:
            save = ''
        for g in self.groups + self.hidden_groups:
            self.fw('module ' + g)
            # --- Check if any variables are derived types. If so, the module
            # --- containing the type must be used.
            printedtypes = []
            for v in self.slist + self.alist:
                if v.group == g and v.derivedtype:
                    if v.type not in printedtypes:
                        self.fw('  use ' + v.type + 'module')
                        printedtypes.append(v.type)
            self.fw('  save')
            # --- Declerations for scalars and arrays
            for s in self.slist:
                if s.group == g:
                    self.fw('  ' + fvars.ftof(s.type), noreturn=1)
                    if s.dynamic:
                        self.fw(', pointer', noreturn=1)
                    self.fw(save + ':: ' + s.name, noreturn=1)
                    if s.data:
                        self.fw('=' + s.data[1:-1], noreturn=1)
                    self.fw('')
            for a in self.alist:
                if a.group == g:
                    if a.dynamic:
                        if a.type == 'character':
                            self.fw('  character(len=' + a.dims[0].high + '), pointer' + save + ':: ' + a.name, noreturn=1)
                            ndims = len(a.dims) - 1
                        else:
                            self.fw('  ' + fvars.ftof(a.type) + ', pointer' + save + ':: ' + a.name, noreturn=1)
                            ndims = len(a.dims)
                        if ndims > 0:
                            self.fw('(' + (ndims*':,')[:-1] + ')', noreturn=1)
                        self.fw('')
                    else:
                        if a.type == 'character':
                            self.fw('  character(len=' + a.dims[0].high + ')' + save + ':: ' + a.name +
                                    re.sub('[ \t\n]', '', a.dimstring))
                        else:
                            self.fw('  ' + fvars.ftof(a.type) + save + ':: ' +
                                    a.name + re.sub('[ \t\n]', '', a.dimstring))
                        if a.data:
                            # --- Add line continuation marks if the data line extends over
                            # --- multiple lines.
                            dd = re.sub(r'\n', '&\n', a.data)
                            self.fw('  data ' + a.name + dd)
            self.fw('end module ' + g)

###############################################################################
###############################################################################
###############################################################################
###############################################################################

module_prefix_pat = re.compile('([a-zA-Z_][a-zA-Z0-9_]*)\.scalars')


def get_another_scalar_dict(file_name, other_scalar_vars):
    m = module_prefix_pat.search(file_name)
    if m.start() == -1:
        raise SyntaxError('expect a .scalars file')
    f = open(file_name, 'rb')
    vars = []
    vars.append(pickle.load(f))
    vars.append(pickle.load(f))
    other_scalar_vars.append(vars)
    f.close()


def wrappergenerator_main(argv=None, writef90modulesonly=0):
    # --- Get package name from argument list
    try:
        pkgname = args.pkgname
        varfile = args.remainder[0]
        otherfortranfiles = args.remainder[1:]
    except IndexError:
        print PyWrap.__doc__
        sys.exit(1)

    # --- get other command line args and default actions
    pkgsuffix = args.pkgsuffix
    pkgbase = args.pkgbase
    initialgallot = args.initialgallot
    fcompname = args.fcomp
    writemodules = args.writemodules
    timeroutines = args.timeroutines
    otherinterfacefiles = args.othermacros

    # --- a list of scalar dictionaries from other modules.
    other_scalar_vars = []
    for d in args.dependencies:
        get_another_scalar_dict(d, other_scalar_vars)

    cc = PyWrap(varfile, pkgname, pkgsuffix, pkgbase, initialgallot, writemodules,
                otherinterfacefiles, other_scalar_vars, timeroutines,
                otherfortranfiles, fcompname)
    if writef90modulesonly:
        cc.writef90modules()
    else:
        cc.createmodulefile()

    # --- forthonf2c.h is imported by Forthon.h, and defines macros needed for strings.
    writeforthonf2c()


# --- This might make some of the write statements cleaner.
# --- From http://aspn.activestate.com/ASPN/Python/Cookbook/
class PrintEval:

    def __init__(self, globals=None, locals=None):
        self.globals = globals or {}
        self.locals = locals or None

    def __getitem__(self, key):
        if self.locals is None:
            self.locals = sys._getframe(1).f_locals
        key = key % self
        return eval(key, self.globals, self.locals)

if __name__ == '__main__':
    wrappergenerator_main(sys.argv[1:])



"""Define Forthon options"""

import sys
import os.path
import argparse
from version import version

parser = argparse.ArgumentParser(
                   usage="Forthon [options] pkgname [extra Fortran or C files to be compiled or objects to link] [options for distutils]",
                   description="""
pkgname is the name of the package.
A complete package will have at least two files, the interface description
file and the fortran file. The default name for the interface file is
pkgname.v. Note that the first line of the interface file must be the package
name. The default name for the fortran file is pkgname.F

Extra files can for fortran or C files that are to be compiled and included
in the package.
"""
)

# --- The argparser stops parsing when it comes across a
# --- non-option argument (i.e. one without the preceeding '-').
# --- This first non-option argument should be the package name and then possibly
# --- extra files to be compiled. Having the parser stop there allows
# --- additional arguments to be added to the command line, after the
# --- extra files, that are passed to distutils during the call to setup.

parser.add_argument('pkgname', help=argparse.SUPPRESS)
parser.add_argument('remainder', nargs=argparse.REMAINDER, help=argparse.SUPPRESS)

parser.add_argument('--build', action='store_true', default=True, dest='dobuild', help='build the package only, the default')
parser.add_argument('--build-base', default='', help='Location where the build related temporary files are put.')
parser.add_argument('--build-temp', default='', help='Location where the *pymodule.o files should be placed. This is relative to the builddir. This defaults to the builddir.')
parser.add_argument('--builddir', help='Location where the temporary compilation files (such as object files) should be placed. This defaults to build/temp-osname.')

parser.add_argument('--cargs', help='Additional options for the C compiler. These are passed through distutils, which does the compilation of C code. If there are any spaces in options, it must be surrounded in double quotes.')
parser.add_argument('--compile_first', default='', metavar="FILE", help='The specified file is compiled first. Normally the file that is compiled first is the fortran file generated by Forthon, which would normally contain all of the modules. But if the modules are in a different file, for example, then that file would need to be compiled first and should be specified here.')

parser.add_argument('-g', '--debug', action='store_true', default=False, help='Turns off optimization for fortran compiler.')
parser.add_argument('-d', '--dependencies', action='append', default=[], help='Specifies that a package that the package being built depends upon. This option can be specified multiple times.')
parser.add_argument('-D', '--defines', action='append', default=[], help='Defines a macro which will be inserted into the makefile. This is required in some cases where a third party library must be specified. This can be specified multiple times.')

parser.add_argument('--f90', action='store_true', default=True, help='Writes wrapper code using f90, which means that python accessible variables are defined in f90 modules. This is the default.')
parser.add_argument('--fargs', action='append', dest='fargslist', default=[], metavar="FARGS", help='Additional options for the fortran compiler. For example to turn on profiling. If there are any spaces in options, it must be surrounded in double quotes.')
parser.add_argument('-F', '--fcomp', help='Fortran compiler. Will automatically be determined if not supplied. It can be one of the following, depending on the machine: intel8, intel, pg, absort, nag, xlf, mpxlf, xlf_r, g95, gfortran.')
parser.add_argument('--fcompexec', help='The executable name of the fortran compiler, if it is different and the compiler name. The -F (--fcomp) option must also be specified.')
parser.add_argument('--with_feenableexcept', action='store_true', default=False, help='Implements feenableexcept call (only supported on Linux)')
parser.add_argument('--fixed_suffix', default='F', help='Suffix to use for fortran files in fixed format. Defaults to F')
parser.add_argument('-f', '--fortranfile', help='Specifiy full name of main fortran file. It defaults to pkgname.F.')
parser.add_argument('--fopt', help='Optimization option for the fortran compiler. This will replace the default optimization options. If there are any spaces in options, it must be surrounded in double quotes.')
parser.add_argument('--free_suffix', default='F90', help='Suffix to use for fortran files in free format. Defaults to F90')

parser.add_argument('--implicitnone', action='store_true', default=True, dest='implicitnone')
parser.add_argument('--noimplicitnone', action='store_false', default=True, dest='implicitnone', help='Specifies whether implicitnone is enforced. The default is --implicitnone.')
parser.add_argument('-I', '--includedirs', action='append', default=[], help='Additional include paths')
parser.add_argument('-a', '--initialgallot', action='store_true', default=False)
parser.add_argument('--noinitialgallot', action='store_false', default=False, help='Specifies whether all groups will be allocated when package is imported into python. The default is --noinitialgallot.')
parser.add_argument('--install', action='store_false', default=True, dest='dobuild', help='Install the package into site-packages')
parser.add_argument('-i', '--interfacefile', help='Specify full name of interface file. It defaults to pkgname.v.')

parser.add_argument('-l', '--libs', action='append', default=[], help="Additional libraries that are needed. Note that the prefix 'lib' and any suffixes should not be included.")
parser.add_argument('-L', '--libdirs', action='append', default=[], help='Additional library paths')

parser.add_argument('-t', '--machine', default=sys.platform, help='Machine type. Will automatically be determined if not supplied. Can be one of linux2, linux3, aix4, aix5, darwin, win32.')
parser.add_argument('--macros', action='append', dest='othermacros', default=[], metavar="MACROS", help='Other interface files whose macros are needed')

parser.add_argument('--pkgbase', default=None, help='Base name of code, when installed. This is use when the package is installed as part of a larger code, when the installed package name is different than the package name of the comiled module.')
parser.add_argument('--pkgdir', default=None, help='Directory where files are that are to be installed with the wrapper')
parser.add_argument('--pkgsuffix', default='', help='Suffix added to the name of the package when installed')

parser.add_argument('--realsize', choices=['4', '8'], default='8', metavar='[4, 8]', help='The size of reals to use for variables that are declared to of type real in the variable description file. It defaults to 8.')

parser.add_argument('--script', action='append', dest='scripts', help='script to be installed in the path. The option can be specified multiple times for multiple scripts.')

parser.add_argument('--timeroutines', action='store_true', default=False)
parser.add_argument('--notimeroutines', action='store_false', default=False, help='Specifies if timers are added for each python callable fortran routine. The default is --notimeroutines.')

parser.add_argument('--underscoring', action='store_true', default=True)
parser.add_argument('--nounderscoring', action='store_false', default=True, dest='underscoring', help='Specifies whether to use any underscores when doing fortran name mangling. For most systems, the default is --underscoring.')
parser.add_argument('--2underscores', action='store_true', default=False, dest='twounderscores')
parser.add_argument('--no2underscores', action='store_false', default=False, dest='twounderscores', help='Specifies whether or not to use second underscores when doing fortran name mangling.')

parser.add_argument('-v', '--verbose', action='store_true', default=False, help='Turn on verbose output during the make process')
parser.add_argument('--version', action='version', version='%(prog)s '+version)

parser.add_argument('--writemodules', action='store_true', default=True, dest='writemodules')
parser.add_argument('--nowritemodules', action='store_false', default=True, dest='writemodules', help="Don't write out the module definitions. Useful if the modules have been written already. Note that if variables of derived type are used, the original code will need to be modified. See example2. Also note that if this option is used, no checks are made to ensure the consistency between the interface file description and the actual module.")

# --- Print help and then exit if no arguments are given
if len(sys.argv) == 1:
    parser.print_help()
    sys.exit(0)

# --- Only process the true argument list when this is called from Forthon.
# --- Otherwise ignore the arguments. This is needed since for example this
# --- module may be imported by the compilers module which is used by some
# --- program other than Forthon.
if os.path.basename(sys.argv[0]).startswith('Forthon') or sys.argv[0] == '-c':
    args = parser.parse_args()
else:
    args = parser.parse_args(args=['dummypkg'])



"""Class which determines which fortran compiler to use and sets defaults
for it.
"""

import sys
import os
import re
import platform
import struct
from cfinterface import realsize, intsize


class FCompiler:
    """
    Determines which compiler to use and sets up description of how to use it.
    To add a new compiler, create a new method with a name using the format
    machine_compiler. The first lines of the function must be of the following form

        if usecompiler(fcompname, fcompexec):
          self.fcompname = fcompname
          self.f90free += ' -fortran arguments for free format'
          self.f90fixed += ' -fortran arguments for fixed format'

    where fcompexec is the executable name of the compiler and fcompname is a
    descriptive (or company) name for the compiler. They can be the same.
    Note that the final linking is done with gcc, so any fortran libraries will
    need to be added to libs (and their locations to libdirs).
    Also, the new function must be included in the while loop below in the
    appropriate block for the machine.
    """

    def __init__(self, machine=None, debug=0, fcompname=None, fcompexec=None, implicitnone=1, twounderscores=0):
        if machine is None:
            machine = sys.platform
        self.machine = machine
        if self.machine != 'win32':
            self.processor = os.uname()[4]
        else:
            self.processor = 'i686'
        self.paths = os.environ['PATH'].split(os.pathsep)

        self.fcompname = fcompname
        self.fcompexec = fcompexec
        self.implicitnone = implicitnone
        self.twounderscores = twounderscores
        self.defines = []
        self.fopt = ''
        self.popt = ''
        self.libs = []
        self.libdirs = []
        self.forthonargs = []
        self.extra_link_args = []
        self.extra_compile_args = []
        self.define_macros = []

        # --- Pick the fortran compiler
        # --- When adding a new compiler, it must be listed here under the correct
        # --- machine name.
        while 1:
            if self.machine in ['linux', 'linux2', 'linux3']:
                if self.linux_intel() is not None:
                    break
                if self.linux_g95() is not None:
                    break
                if self.linux_gfortran() is not None:
                    break
                if self.linux_pg() is not None:
                    break
                if self.linux_absoft() is not None:
                    break
                if self.linux_lahey() is not None:
                    break
                if self.linux_pathscale() is not None:
                    break
                if self.linux_xlf_r() is not None:
                    break
                if self.linux_cray() is not None:
                    break
            elif self.machine == 'darwin':
                if self.macosx_xlf() is not None:
                    break
                if self.macosx_g95() is not None:
                    break
                if self.macosx_gfortran() is not None:
                    break
                if self.macosx_absoft() is not None:
                    break
                if self.macosx_nag() is not None:
                    break
                if self.macosx_gnu() is not None:
                    break
            elif self.machine == 'cygwin':
                if self.cygwin_g95() is not None:
                    break
            elif self.machine == 'win32':
                if self.win32_pg() is not None:
                    break
                if self.win32_intel() is not None:
                    break
            elif self.machine == 'aix4' or self.machine == 'aix5':
                if self.aix_xlf() is not None:
                    break
                if self.aix_mpxlf() is not None:
                    break
                if self.aix_xlf_r() is not None:
                    break
                if self.aix_mpxlf64() is not None:
                    break
                if self.aix_pghpf() is not None:
                    break
            else:
                raise SystemExit('Machine type %s is unknown'%self.machine)
            raise SystemExit('Fortran compiler not found')

        # --- The following two quantities must be defined.
        try:
            self.f90free
            self.f90fixed
        except:
            # --- Note that this error should never happed (except during debugging)
            raise ValueError("The fortran compiler definition is not correct, f90free and f90fixed must be defined.")

        if debug:
            self.fopt = '-g'
            self.popt = '-g'
            self.extra_link_args += ['-g']
            self.extra_compile_args += ['-g', '-O0']

        # --- Add the compiler name to the forthon arguments
        self.forthonargs += ['-F ' + self.fcompname]

    def usecompiler(self, fcompname, fcompexec):
        'Check if the specified compiler is found'
        if self.fcompexec is None:
            result = (self.findfile(fcompexec) and
                      (self.fcompname == fcompname or self.fcompname is None))
            if result:
                self.fcompexec = fcompexec
        else:
            result = self.fcompname == fcompname and self.findfile(self.fcompexec)
        if result:
            self.f90free = self.fcompexec
            self.f90fixed = self.fcompexec
        return result

    def findfile(self, file, followlinks=1):
        if self.machine == 'win32':
            file = file + '.exe'
        if os.path.isabs(file):
            if os.access(file, os.X_OK):
                return file
            return None
        for path in self.paths:
            try:
                if file in os.listdir(path) and os.path.isfile(os.path.join(path, file)):
                    # --- Check if the path is a link
                    if followlinks:
                        try:
                            link = os.readlink(os.path.join(path, file))
                            result = os.path.dirname(link)
                            if result == '':
                                # --- link is not a full path but a local link, so the
                                # --- path needs to be prepended.
                                result = os.path.join(os.path.dirname(path), link)
                            path = result
                        except OSError:
                            pass
                    return path
            except:
                pass
        return None

    # ----------------------------------------------------------------------------
    # --- Machine generic utilities

    # --- For g95 and gfortran
    def findgnulibroot(self, fcompname, fcompexec):
        # --- Find the lib root for gnu based compilers.
        # --- Get the full name of the compiler executable.
        fcomp = os.path.join(self.findfile(fcompexec, followlinks=0), fcompexec)
        # --- Map the compiler name to the library needed.
        flib = {'gfortran': 'gfortran', 'g95': 'f95'}[fcompname]
        # --- Run it with the appropriate option to return the library path name
        ff = os.popen(fcomp + ' -print-file-name=lib' + flib + '.a')
        gcclib = ff.readline()[:-1]
        ff.close()
        # --- Strip off the actual library name to get the path.
        libroot = os.path.dirname(gcclib)
        return libroot

    # --- Makes sure libdirs is not [''], because that can change the semantics
    # --- of how gcc linking works. A -L'' argument will cause global libraries
    # --- to not be found.
    def findgnulibdirs(self, fcompname, fcompexec):
        libroot = self.findgnulibroot(fcompname, fcompexec)
        if libroot:
            return [libroot]
        return []

    # --- Make sure tha the version of gfortran is new enough. Older versions
    # --- had a bug - array dimensions were not setup properly for array arguments,
    # --- leading to memory out of bounds errors.
    def isgfortranversionok(self, fcompexec):
        fcomp = os.path.join(self.findfile(fcompexec, followlinks=0), fcompexec)
        ff = os.popen(fcomp + ' -dumpversion')
        dumpversion = ff.readline()[:-1]
        ff.close()
        # --- The format of dumpversion is not consistent - really annoying.
        # --- For version 4.4 and older, there was extra text output.
        # --- Search through it to find a version number.
        for version in dumpversion.split():
            if re.match('[0-9]', version[0]):
                break
        else:
            # --- Version not found, set to zero
            version = '0.0'
        vfloat = float('.'.join(version.split('.')[0:2]))
        if vfloat < 4.3:
            print "gfortran will not be used, it's version is too old or unknown - upgrade to a newer version"
            return False
        else:
            return True

    # -----------------------------------------------------------------------------
    # --- LINUX
    def linux_intel(self):
        if self.usecompiler('intel', 'ifort') or self.usecompiler('intel8', 'ifort'):
            self.fcompname = 'ifort'
            self.f90free += ' -nofor_main -free -DIFC -fpp -fPIC'
            self.f90fixed += ' -nofor_main -132 -DIFC -fpp -fPIC'
            self.f90free += ' -DFPSIZE=%s -r%s -Zp%s'%(realsize, realsize, realsize)
            self.f90fixed += ' -DFPSIZE=%s -r%s -Zp%s'%(realsize, realsize, realsize)
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            if self.implicitnone:
                self.f90free += ' -implicitnone'
                self.f90fixed += ' -implicitnone'
            self.popt = '-O'
            flibroot, b = os.path.split(self.findfile('ifort'))
            self.libdirs = [flibroot + '/lib']
            self.libs = ['ifcore', 'ifport', 'imf', 'svml', 'irc']
            cpuinfo = open('/proc/cpuinfo', 'r').read()
            if re.search('Pentium III', cpuinfo):
                self.fopt = '-O3 -xK -tpp6 -ip -unroll -prefetch'
            elif re.search('AMD Athlon', cpuinfo):
                self.fopt = '-O3 -ip -unroll -prefetch'
            elif self.processor == 'ia64':
                self.fopt = '-O3 -ip -unroll -tpp2'
                self.f90free += ' -fpic'
                self.f90fixed += ' -fpic'
                self.libs.remove('svml')
            elif struct.calcsize('l') == 8:
                self.fopt = '-O3 -ip -unroll'
            else:
                self.fopt = '-O3 -ip -unroll'
            return 1

    def linux_g95(self):
        if self.usecompiler('g95', 'g95'):
            self.fcompname = 'g95'
            self.f90free += ' -ffree-form -fPIC -Wno=155 -fshort-circuit'
            self.f90fixed += ' -ffixed-line-length-132 -fPIC -fshort-circuit'
            self.f90free += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90fixed += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            if self.implicitnone:
                self.f90free += ' -fimplicit-none'
                self.f90fixed += ' -fimplicit-none'
            if self.twounderscores:
                self.f90free += ' -fsecond-underscore'
                self.f90fixed += ' -fsecond-underscore'
            else:
                self.f90free += ' -fno-second-underscore'
                self.f90fixed += ' -fno-second-underscore'
            self.popt = '-O'
            self.libdirs = self.findgnulibdirs('g95', self.fcompexec)
            self.libs = ['f95']
            cpuinfo = open('/proc/cpuinfo', 'r').read()
            if re.search('Pentium III', cpuinfo):
                self.fopt = '-O3'
            elif re.search('AMD Athlon', cpuinfo):
                self.fopt = '-O3'
            elif struct.calcsize('l') == 8:
                self.fopt = '-O3 -mfpmath=sse -ftree-vectorize -ftree-vectorizer-verbose=0 -funroll-loops -fstrict-aliasing -fsched-interblock -falign-loops=16 -falign-jumps=16 -falign-functions=16 -ffast-math -fstrict-aliasing'
            else:
                self.fopt = '-O3'
            return 1

    def linux_gfortran(self):
        if self.usecompiler('gfortran', 'gfortran'):
            if not self.isgfortranversionok(self.fcompexec):
                return None
            self.fcompname = 'gfortran'
            self.f90free += ' -fPIC'
            self.f90fixed += ' -fPIC -ffixed-line-length-132'
            self.f90free += ' -DFPSIZE=%s'%(realsize)
            self.f90fixed += ' -DFPSIZE=%s'%(realsize)
            if realsize == '8':
                self.f90free += ' -fdefault-real-8 -fdefault-double-8'
                self.f90fixed += ' -fdefault-real-8 -fdefault-double-8'
            self.f90free += ' -DISZ=%s'%(intsize)
            self.f90fixed += ' -DISZ=%s'%(intsize)
            if intsize == '8':
                self.f90free += ' -fdefault-integer-8'
                self.f90fixed += ' -fdefault-integer-8'
            if self.implicitnone:
                self.f90free += ' -fimplicit-none'
                self.f90fixed += ' -fimplicit-none'
            if self.twounderscores:
                self.f90free += ' -fsecond-underscore'
                self.f90fixed += ' -fsecond-underscore'
            else:
                self.f90free += ' -fno-second-underscore'
                self.f90fixed += ' -fno-second-underscore'
            self.libdirs = self.findgnulibdirs('gfortran', self.fcompexec)
            self.libs = ['gfortran']
            self.fopt = '-O3 -ftree-vectorize -ftree-vectorizer-verbose=0'
            return 1

    def linux_pg(self):
        if self.usecompiler('pg', 'pgf90'):
            # --- Portland group
            self.fcompname = 'pgi'
            f90opts = ' -fPIC -Mextend'
            if self.implicitnone:
                f90opts += ' -Mdclchk'
            else:
                f90opts += ' -Mnodclchk'
            if self.twounderscores:
                f90opts += ' -Msecond_underscore'
            else:
                f90opts += ' -Mnosecond_underscore'
            f90opts += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            f90opts += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90free += f90opts
            self.f90fixed += f90opts
            self.popt = '-Mcache_align'
            if platform.python_compiler().startswith('GCC'):
                flibroot, b = os.path.split(self.findfile('pgf90'))
                self.libdirs = [flibroot + '/lib']
                self.libs = ['pgf90']  # ???
            else:
                # --- When using pgcc for linking, this includes the needed fortran libraries.
                self.extra_link_args = ['-pgf90libs']
            self.fopt = '-fast -Mcache_align'
            return 1

    def linux_absoft(self):
        if self.usecompiler('absoft', 'f90'):
            self.fcompname = 'absoft'
            # --- Absoft
            self.f90free += ' -B108 -N113 -W132 -YCFRL=1 -YEXT_NAMES=ASIS'
            self.f90fixed += ' -B108 -N113 -W132 -YCFRL=1 -YEXT_NAMES=ASIS'
            self.f90free += ' -DFPSIZE=%s'%(realsize)  # ???
            self.f90fixed += ' -DFPSIZE=%s'%(realsize)  # ???
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.popt = '-Mcache_align'
            self.forthonargs = ['--2underscores']  # --- This needs to be fixed XXX
            flibroot, b = os.path.split(self.findfile('f90'))
            self.libdirs = [flibroot + '/lib']
            self.libs = ['U77', 'V77', 'f77math', 'f90math', 'fio']
            self.fopt = '-O'
            return 1

    def linux_lahey(self):
        if self.usecompiler('lahey', 'lf95'):
            self.fcompname = 'lahey'
            # --- Lahey
            # in = implicit none
            # dbl = real*8 (variables or constants?)
            # [n]fix = fixed or free form
            # wide = column width longer than 72
            # ap = preserve arithmetic precision
            self.f90free += ' --nfix --dbl --mlcdecl'
            self.f90fixed += ' --fix --wide --dbl --mlcdecl'
            self.f90free += ' -DFPSIZE=%s'%(realsize)  # ???
            self.f90fixed += ' -DFPSIZE=%s'%(realsize)  # ???
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.popt = '-Mcache_align'
            if self.implicitnone:
                self.f90free += ' --in'
                self.f90fixed += ' --in'
            self.popt = '-O'
            flibroot, b = os.path.split(self.findfile('lf95'))
            self.libdirs = [flibroot + '/lib']
            self.libs = ["fj9i6", "fj9f6", "fj9e6", "fccx86"]
            cpuinfo = open('/proc/cpuinfo', 'r').read()
            if re.search('Pentium III', cpuinfo):
                self.fopt = '--O2 --unroll --prefetch --nap --npca --ntrace --nsav'
            elif re.search('AMD Athlon', cpuinfo):
                self.fopt = '--O2 --unroll --prefetch --nap --npca --ntrace --nsav'
            else:
                self.fopt = '--O2 --unroll --prefetch --nap --npca --ntrace --nsav'
            return 1

    def linux_pathscale(self):
        if self.usecompiler('pathscale', 'pathf95'):
            self.f90free += ' -freeform -DPATHF90 -ftpp -fPIC -woff1615'
            self.f90fixed += ' -fixedform -extend_source -DPATHF90 -ftpp -fPIC -woff1615'
            self.f90free += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90fixed += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            if self.twounderscores:
                self.f90free += ' -fsecond-underscore'
                self.f90fixed += ' -fsecond-underscore'
            else:
                self.f90free += ' -fno-second-underscore'
                self.f90fixed += ' -fno-second-underscore'
            self.popt = '-O'
            flibroot, b = os.path.split(self.findfile(self.fcompname))
            self.libdirs = [flibroot + '/lib/2.1']
            self.libs = ['pathfortran']
            cpuinfo = open('/proc/cpuinfo', 'r').read()
            self.extra_compile_args = ['-fPIC']
            self.extra_link_args = ['-fPIC']
            if re.search('Pentium III', cpuinfo):
                self.fopt = '-Ofast'
            elif re.search('AMD Athlon', cpuinfo):
                self.fopt = '-O3'
            elif struct.calcsize('l') == 8:
                self.fopt = '-O3 -OPT:Ofast -fno-math-errno'
            else:
                self.fopt = '-Ofast'
            return 1

    def linux_xlf_r(self):
        if self.usecompiler('xlf_r', 'xlf95_r'):
            self.fcompname = 'xlf'
            intsize = struct.calcsize('l')
            f90 = ' -c -WF,-DXLF -qmaxmem=8192 -qdpc=e -qautodbl=dbl4 -qsave=defaultinit -WF,-DESSL'%locals()
            self.f90free += f90 + ' -qsuffix=f=f90:cpp=F90 -qfree=f90'
            self.f90fixed += f90 + ' -qfixed=132'
            self.f90free += ' -WF,-DFPSIZE=%s'%(realsize)  # ???
            self.f90fixed += ' -WF,-DFPSIZE=%s'%(realsize)  # ???
            self.f90free += ' -WF,-DISZ=%s -qintsize%s'%(intsize, intsize)
            self.f90fixed += ' -WF,-DISZ=%s -qintsize%s'%(intsize, intsize)
            self.ld = 'xlf95_r -bE:$(PYTHON)/lib/python$(PYVERS)/config/python.exp'%locals()
            if self.implicitnone:
                self.f90free += ' -u'
                self.f90fixed += ' -u'
            self.popt = '-O'
            self.extra_link_args = []
            self.extra_compile_args = []
            # --- Note that these are specific the machine intrepid at Argonne.
            self.libdirs = ['/gpfs/software/linux-sles10-ppc64/apps/V1R3M0/ibmcmp-sep2008/opt/xlf/bg/11.1/lib', '/gpfs/software/linux-sles10-ppc64/apps/V1R3M0/ibmcmp-sep2008/opt/xlsmp/bg/1.7/lib']
            self.libs = ['xlf90_r', 'xlsmp']
            self.fopt = '-O3 -qstrict -qarch=auto -qtune=auto -qsmp=omp'
            return 1

    def linux_cray(self):
        if self.usecompiler('crayftn', 'crayftn'):
            self.fcompname = 'crayftn'
            self.f90free += ' -f free'
            self.f90fixed += ' -f fixed -N 132'
            self.f90free += ' -DFPSIZE=%s'%(realsize)
            self.f90fixed += ' -DFPSIZE=%s'%(realsize)
            if realsize == '8':
                self.f90free += ' -s real64'
                self.f90fixed += ' -s real64'
            self.f90free += ' -DISZ=%s'%(intsize)
            self.f90fixed += ' -DISZ=%s'%(intsize)
            if intsize == '8':
                self.f90free += ' -s integer64'
                self.f90fixed += ' -s integer64'
            if self.implicitnone:
                self.f90free += ' -e I'
                self.f90fixed += ' -e I'
            if self.twounderscores:
                self.f90free += ' -h second_underscore'
                self.f90fixed += ' -h second_underscore'
            else:
                self.f90free += ' -h nosecond_underscore'
                self.f90fixed += ' -h nosecond_underscore'
            self.libdirs = []
            self.libs = []
            self.fopt = '-O3'
            return 1

    # -----------------------------------------------------------------------------
    # --- CYGWIN
    def cygwin_g95(self):
        if self.usecompiler('g95', 'g95'):
            self.fcompname = 'g95'
            self.f90fixed += ' -ffixed-line-length-132'
            self.f90free += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90fixed += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            if self.twounderscores:
                self.f90free += ' -fsecond-underscore'
                self.f90fixed += ' -fsecond-underscore'
            else:
                self.f90free += ' -fno-second-underscore'
                self.f90fixed += ' -fno-second-underscore'
            self.fopt = '-O3 -ftree-vectorize -ftree-vectorizer-verbose=0'
#      self.fopt = '-O3 -funroll-loops -fstrict-aliasing -fsched-interblock  \
#           -falign-loops=16 -falign-jumps=16 -falign-functions=16 \
#           -falign-jumps-max-skip=15 -falign-loops-max-skip=15 -malign-natural \
#           -ffast-math -mpowerpc-gpopt -force_cpusubtype_ALL \
#           -fstrict-aliasing'
#      self.extra_link_args = ['-flat_namespace', '-undefined suppress', '-lg2c']
            self.extra_link_args = ['-flat_namespace', '--allow-shlib-undefined', '-Wl,--export-all-symbols', '-Wl,-export-dynamic', '-Wl,--unresolved-symbols=ignore-all', '-lg2c']
            self.libdirs = self.findgnulibdirs('g95', self.fcompexec)
            self.libdirs.append('/lib/w32api')
            self.libs = ['f95']
            return 1

    # -----------------------------------------------------------------------------
    # --- MAC OSX
    def macosx_g95(self):
        if self.usecompiler('g95', 'g95'):
            self.fcompname = 'g95'
            self.f90free += ' -fzero -ffree-form -Wno=155'
            self.f90fixed += ' -fzero -ffixed-line-length-132'
            self.f90free += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90fixed += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            if self.implicitnone:
                self.f90free += ' -fimplicit-none'
                self.f90fixed += ' -fimplicit-none'
            if self.twounderscores:
                self.f90free += ' -fsecond-underscore'
                self.f90fixed += ' -fsecond-underscore'
            else:
                self.f90free += ' -fno-second-underscore'
                self.f90fixed += ' -fno-second-underscore'
            self.fopt = '-O3 -funroll-loops -fstrict-aliasing -fsched-interblock \
                 -falign-loops=16 -falign-jumps=16 -falign-functions=16 \
                 -ftree-vectorize -ftree-vectorizer-verbose=0 \
                 -ffast-math -fstrict-aliasing'
#      self.fopt = '-O3  -mtune=G5 -mcpu=G5 -mpowerpc64'
            self.extra_link_args = ['-flat_namespace']
            self.libdirs = self.findgnulibdirs('g95', self.fcompexec)
            self.libs = ['f95']
            return 1

    def macosx_gfortran(self):
        if self.usecompiler('gfortran', 'gfortran'):
            if not self.isgfortranversionok(self.fcompexec):
                return None
            self.fcompname = 'gfortran'
            self.f90fixed += ' -ffixed-line-length-132'
            self.f90free += ' -DFPSIZE=%s'%(realsize)
            self.f90fixed += ' -DFPSIZE=%s'%(realsize)
            if realsize == '8':
                self.f90free += ' -fdefault-real-8 -fdefault-double-8'
                self.f90fixed += ' -fdefault-real-8 -fdefault-double-8'
            self.f90free += ' -DISZ=%s'%(intsize)
            self.f90fixed += ' -DISZ=%s'%(intsize)
            if intsize == '8':
                self.f90free += ' -fdefault-integer-8'
                self.f90fixed += ' -fdefault-integer-8'
            if self.implicitnone:
                self.f90free += ' -fimplicit-none'
                self.f90fixed += ' -fimplicit-none'
            if self.twounderscores:
                self.f90free += ' -fsecond-underscore'
                self.f90fixed += ' -fsecond-underscore'
            else:
                self.f90free += ' -fno-second-underscore'
                self.f90fixed += ' -fno-second-underscore'
            flibroot, b = os.path.split(self.findfile('gfortran'))
            self.fopt = '-O3 -funroll-loops -fstrict-aliasing -fsched-interblock  \
                 -falign-loops=16 -falign-jumps=16 -falign-functions=16 \
                 -malign-natural \
                 -ffast-math -mpowerpc-gpopt -force_cpusubtype_ALL \
                 -fstrict-aliasing -mtune=G5 -mcpu=G5 -mpowerpc64'
#      self.fopt = '-O3  -mtune=G5 -mcpu=G5 -mpowerpc64'
            self.fopt = '-O3 -ftree-vectorize -ftree-vectorizer-verbose=0'
#      self.extra_link_args = ['-flat_namespace', '-lg2c']
            self.extra_link_args = ['-flat_namespace']
            self.libdirs = self.findgnulibdirs('gfortran', self.fcompexec)
            self.libs = ['gfortran']
            return 1

    def macosx_xlf(self):
        if self.usecompiler('xlf', 'xlf95') or self.usecompiler('xlf90', 'xlf95'):
            self.fcompname = 'xlf'
            self.f90free += ' -WF,-DXLF -qsuffix=f=f90:cpp=F90 -qextname -qautodbl=dbl4 -qdpc=e -bmaxdata:0x70000000 -bmaxstack:0x10000000 -qinitauto'
            self.f90fixed += ' -WF,-DXLF -qextname -qfixed=132 -qautodbl=dbl4 -qdpc=e -bmaxdata:0x70000000 -bmaxstack:0x10000000 -qinitauto'
            self.f90free += ' -WF,-DFPSIZE=%s'%(realsize)  # ???
            self.f90fixed += ' -WF,-DFPSIZE=%s'%(realsize)  # ???
            self.f90free += ' -WF,-DISZ=%s -qintsize%s'%(intsize, intsize)
            self.f90fixed += ' -WF,-DISZ=%s -qintsize%s'%(intsize, intsize)
            if self.implicitnone:
                self.f90free += ' -u'
                self.f90fixed += ' -u'
            self.fopt = '-O5'
            self.extra_link_args = ['-flat_namespace']  # , '-Wl,-undefined, suppress']  # , '-Wl,-stack_size, 10000000']
            flibroot, b = os.path.split(self.findfile('xlf95'))
            self.libdirs = [flibroot + '/lib']
            self.libs = ['xlf90', 'xl', 'xlfmath']
            return 1

    def macosx_absoft(self):
        if self.usecompiler('absoft', 'f90'):
            self.fcompname = 'absoft'
            self.f90free += ' -N11 -N113 -YEXT_NAMES=LCS -YEXT_SFX=_'
            self.f90fixed += ' -f fixed -W 132 -N11 -N113 -YEXT_NAMES=LCS -YEXT_SFX=_'
            self.f90free += ' -DFPSIZE=%s'%(realsize)  # ???
            self.f90fixed += ' -DFPSIZE=%s'%(realsize)  # ???
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            flibroot, b = os.path.split(self.findfile('f90'))
            self.libdirs = [flibroot + '/lib']
            self.extra_link_args = ['-flat_namespace', '-Wl,-undefined, suppress']
            self.libs = ['fio', 'f77math', 'f90math', 'f90math_altivec', 'lapack', 'blas']
            self.fopt = '-O3'
            return 1

    def macosx_nag(self):
        if self.usecompiler('nag', 'f95'):
            self.fcompname = 'nag'
            # self.f90free += ' -132 -fpp -Wp,-macro=no_com -Wc,-O3 -Wc,-funroll-loops -free -PIC -u -w -mismatch_all -kind=byte'
            # self.f90fixed += ' -132 -fpp -u -Wp,-macro=no_com -Wp,-fixed -fixed -Wc,-O3 -Wc,-funroll-loops -PIC -w -mismatch_all -kind=byte'
            self.f90free += ' -132 -fpp -Wp,-macro=no_com -free -PIC -u -w -mismatch_all -kind=byte -Oassumed=contig'
            self.f90fixed += ' -132 -fpp -Wp,-macro=no_com -Wp,-fixed -fixed -PIC -u -w -mismatch_all -kind=byte -Oassumed=contig'
            self.f90free += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90fixed += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            flibroot, b = os.path.split(self.findfile('f95'))
            self.extra_link_args = ['-flat_namespace', '-framework vecLib', '/usr/local/lib/NAGWare/quickfit.o', '/usr/local/lib/NAGWare/libf96.a']
            self.libs = ['m']
            self.fopt = '-Wc,-O3 -Wc,-funroll-loops -O3 -Ounroll=2'
            self.fopt = '-O4 -Wc,-fast'
            self.fopt = '-O3 '  # -Wc,-fast'
            self.define_macros.append(('NAG', '1'))
            return 1

    def macosx_gnu(self):
        if self.usecompiler('gnu', 'g95'):
            self.fcompname = 'gnu'
            self.f90fixed += ' -ffixed-form -ffixed-line-length-132'
            self.f90free += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90fixed += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            flibroot, b = os.path.split(self.findfile('g95'))
            self.libdirs = [flibroot + '/lib']
            self.libs = ['???']
            self.fopts = '-O3'
            return 1

    # -----------------------------------------------------------------------------
    # --- WIN32
    def win32_pg(self):
        if self.usecompiler('pg', 'pgf90'):
            # --- Portland group
            self.fcompname = 'pgi'
            self.f90free += ' -Mextend -Mdclchk'
            self.f90fixed += ' -Mextend -Mdclchk'
            self.f90free += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90fixed += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.popt = '-Mcache_align'
            flibroot, b = os.path.split(self.findfile('pgf90'))
            self.libdirs = [flibroot + '/Lib']
            self.libs = ['???']
            self.fopt = '-fast -Mcache_align'
            return 1

    def win32_intel(self):
        if self.usecompiler('intel', 'ifl'):
            self.fcompname = 'ifl'
            self.f90free += ' -Qextend_source -Qautodouble -DIFC -FR -Qfpp -4Yd -C90 -Zp8 -Qlowercase -us -MT -Zl -static'
            self.f90fixed += ' -Qextend_source -Qautodouble -DIFC -FI -Qfpp -4Yd -C90 -Zp8 -Qlowercase -us -MT -Zl -static'
            self.f90free += ' -DFPSIZE=%s'%(realsize)  # ???
            self.f90fixed += ' -DFPSIZE=%s'%(realsize)  # ???
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            flibroot, b = os.path.split(self.findfile('ifl'))
            self.libdirs = [flibroot + '/Lib']
            self.libs = ['CEPCF90MD', 'F90MD', 'intrinsMD']
            self.fopt = '-O3'
            return 1

    # -----------------------------------------------------------------------------
    # --- AIX
    def aix_xlf(self):
        if self.usecompiler('xlf', 'xlf95'):
            self.fcompname = 'xlf'
            intsize = struct.calcsize('l')
            if intsize == '4':
                bmax = '-bmaxdata:0x70000000 -bmaxstack:0x10000000'
            else:
                bmax = '-q64'
            f90 = ' -c -WF,-DXLF -qmaxmem=8192 -qdpc=e -qautodbl=dbl4 -qsave=defaultinit -WF,-DESSL %(bmax)s'%locals()
            self.f90free += f90 + ' -qsuffix=f=f90:cpp=F90 -qfree=f90'
            self.f90fixed += f90 + ' -qfixed=132'
            self.f90free += ' -WF,-DFPSIZE=%s'%(realsize)  # ???
            self.f90fixed += ' -WF,-DFPSIZE=%s'%(realsize)  # ???
            self.f90free += ' -WF,-DISZ=%s -qintsize%s'%(intsize, intsize)
            self.f90fixed += ' -WF,-DISZ=%s -qintsize%s'%(intsize, intsize)
            self.ld = 'xlf -bE:$(PYTHON)/lib/python$(PYVERS)/config/python.exp %(bmax)s'%locals()
            self.popt = '-O'
            self.extra_link_args = [bmax]
            self.extra_compile_args = [bmax]
            if self.implicitnone:
                self.f90free += ' -u'
                self.f90fixed += ' -u'
            self.libs = ['xlf90', 'xlopt', 'xlf', 'xlomp_ser', 'pthread', 'essl']
            self.fopt = '-O3 -qstrict -qarch=auto -qtune=auto'
            return 1

    def aix_mpxlf(self):
        if self.usecompiler('mpxlf', 'mpxlf95'):
            self.fcompname = 'xlf'
            intsize = struct.calcsize('l')
            if intsize == '4':
                bmax = '-bmaxdata:0x70000000 -bmaxstack:0x10000000'
            else:
                bmax = '-q64'
            f90 = ' -c -WF,-DXLF -qmaxmem=8192 -qdpc=e -qautodbl=dbl4 -qsave=defaultinit -WF,-DMPIPARALLEL -WF,-DESSL %(bmax)s'%locals()
            self.f90free += f90 + ' -qsuffix=f=f90:cpp=F90 -qfree=f90'
            self.f90fixed += f90 + ' -qfixed=132'
            self.f90free += ' -WF,-DFPSIZE=%s'%(realsize)  # ???
            self.f90fixed += ' -WF,-DFPSIZE=%s'%(realsize)  # ???
            self.f90free += ' -WF,-DISZ=%s -qintsize%s'%(intsize, intsize)
            self.f90fixed += ' -WF,-DISZ=%s -qintsize%s'%(intsize, intsize)
            self.ld = 'mpxlf_r -bE:$(PYTHON)/lib/python$(PYVERS)/config/python.exp %(bmax)s'%locals()
            if self.implicitnone:
                self.f90free += ' -u'
                self.f90fixed += ' -u'
            self.popt = '-O'
            self.extra_link_args = [bmax]
            self.extra_compile_args = [bmax]
            self.libs = ['xlf90', 'xlopt', 'xlf', 'xlomp_ser', 'pthread', 'essl']
            self.defines = ['PYMPI=/usr/common/homes/g/grote/pyMPI']
            self.fopt = '-O3 -qstrict -qarch=auto -qtune=auto'
            return 1

    def aix_xlf_r(self):
        if self.usecompiler('xlf_r', 'xlf95_r'):
            # --- IBM SP, OpenMP
            self.fcompname = 'xlf'
            intsize = struct.calcsize('l')
            if intsize == '4':
                bmax = '-bmaxdata:0x70000000 -bmaxstack:0x10000000'
            else:
                bmax = '-q64'
            f90 = ' -c -WF,-DXLF -qmaxmem=8192 -qdpc=e -qautodbl=dbl4 -qsave=defaultinit -WF,-DESSL %(bmax)s'%locals()
            self.f90free += f90 + ' -qsuffix=f=f90:cpp=F90 -qfree=f90'
            self.f90fixed += f90 + ' -qfixed=132'
            self.f90free += ' -WF,-DFPSIZE=%s'%(realsize)  # ???
            self.f90fixed += ' -WF,-DFPSIZE=%s'%(realsize)  # ???
            self.f90free += ' -WF,-DISZ=%s -qintsize%s'%(intsize, intsize)
            self.f90fixed += ' -WF,-DISZ=%s -qintsize%s'%(intsize, intsize)
            self.ld = 'xlf95_r -bE:$(PYTHON)/lib/python$(PYVERS)/config/python.exp %(bmax)s'%locals()
            if self.implicitnone:
                self.f90free += ' -u'
                self.f90fixed += ' -u'
            self.popt = '-O'
            self.extra_link_args = [bmax]
            self.extra_compile_args = [bmax]
            self.libs = ['xlf90', 'xlopt', 'xlf', 'xlsmp', 'pthreads', 'essl']
            self.fopt = '-O3 -qstrict -qarch=auto -qtune=auto -qsmp=omp'
            return 1

    def aix_mpxlf64(self):
        if self.usecompiler('mpxlf64', 'mpxlf95'):
            # --- IBM SP, parallel
            self.fcompname = 'xlf'
            intsize = struct.calcsize('l')
            if intsize == '4':
                bmax = '-bmaxdata:0x70000000 -bmaxstack:0x10000000'
            else:
                bmax = '-q64'
            f90 = ' -c -WF,-DXLF -qmaxmem=8192 -qdpc=e -qautodbl=dbl4 -qsave=defaultinit -WF,-DMPIPARALLEL -WF,-DESSL %(bmax)s'%locals()
            self.f90free += f90 + ' -qsuffix=f=f90:cpp=F90 -qfree=f90'
            self.f90fixed += f90 + ' -qfixed=132'
            self.f90free += ' -WF,-DFPSIZE=%s'%(realsize)  # ???
            self.f90fixed += ' -WF,-DFPSIZE=%s'%(realsize)  # ???
            self.f90free += ' -WF,-DISZ=%s -qintsize%s'%(intsize, intsize)
            self.f90fixed += ' -WF,-DISZ=%s -qintsize%s'%(intsize, intsize)
            self.ld = 'mpxlf95_r -bE:$(PYTHON)/lib/python$(PYVERS)/config/python.exp %(bmax)s'%locals()
            if self.implicitnone:
                self.f90free += ' -u'
                self.f90fixed += ' -u'
            self.popt = '-O'
            self.extra_link_args = [bmax]
            self.extra_compile_args = [bmax]
            self.libs = ['xlf90', 'xlopt', 'xlf', 'xlomp_ser', 'pthread', 'essl']
            self.defines = ['PYMPI=/usr/common/homes/g/grote/pyMPI']
            self.fopt = '-O3 -qstrict -qarch=auto -qtune=auto'
            return 1

    def aix_pghpf(self):
        if self.usecompiler('pghpf', 'pghpf'):
            # --- Portland group
            self.fcompname = 'pghpf'
            self.f90free += ' -Mextend -Mdclchk'
            self.f90fixed += ' -Mextend -Mdclchk'
            self.f90free += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90fixed += ' -DFPSIZE=%s -r%s'%(realsize, realsize)
            self.f90free += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.f90fixed += ' -DISZ=%s -i%s'%(intsize, intsize)
            self.popt = '-Mcache_align'
            flibroot, b = os.path.split(self.findfile('pghpf'))
            self.libdirs = [flibroot + '/lib']
            self.libs = ['pghpf']  # ???
            self.fopt = '-fast -Mcache_align'
            return 1



# Created by David P. Grote, March 6, 1998

import cfinterface

# Declare a class to hold fortran variable info
class Fvars:
    name = ''
    type = ''
    dims = []
    args = []
    dynamic = 0
    data = ''
    unit = ''
    comment = ''
    group = ''
    attr = ''
    limit = ''
    dimstring = ''
    array = 0
    function = 0
    derivedtype = 0
    parameter = 0
    setaction = None
    getaction = None

class Fargs:
    name = ''
    type = ''
    dimstring = ''
    dims = []

class Fdims:
    low = ''
    high = ''

class Ftype:
    def __init__(s, name, attr):
        s.name = name
        s.attr = attr
        s.vlist = []
    def addvar(s, v):
        s.vlist.append(v)
    def display(s):
        print 'name = ', s.name
        print 'variables:'
        for v in s.vlist:
            print '  ', v.name, '  ', v.type

ftoc_dict = {'integer':'long', 'logical':'long',
             'real':{'8':'double', '4':'float'}[cfinterface.realsize],
             'double':'double', 'float':'float',
             'character':'FSTRING', 'string':'FSTRING',
             'void':'void', 'Filedes':'long', 'complex':'Py_complex'}
ftop_dict = {'integer':'LONG', 'logical':'LONG',
             'real':{'8':'DOUBLE', '4':'FLOAT'}[cfinterface.realsize],
             'double':'DOUBLE', 'float':'FLOAT',
             'character':'STRING', 'string':'STRING', 'void':'VOID',
             'Filedes':'LONG', 'complex':'CDOUBLE'}
fto1 = {'integer':'l', 'logical':'l',
        'real':{'8':'d', '4':'f'}[cfinterface.realsize],
        'double':'d', 'float':'f',
        'character':'s',
        'string':'s', 'Filedes':'l', 'complex':'D'}
ftof_dict = {'integer':'integer('+cfinterface.isz+')',
             'real':'real(kind=%s)'%cfinterface.realsize,
             'double':'real(kind=8)',
             'float':'real(kind=4)',
             'logical':'logical('+cfinterface.isz+')',
             'character':'character',
             'string':'character',
             'void':'void',
             'Filedes':'integer('+cfinterface.isz+')',
             'complex':'complex(kind=8)'}

def isderivedtype(arg):
    if arg.type in ftoc_dict: return 0
    else: return 1

def ftoc(type):
    if type in ftoc_dict: return ftoc_dict[type]
    else: return 'char'

def ftop(type):
    # --- Returns the numpy type associated with the fortran type.
    try:
        result = ftop_dict[type]
    except KeyError:
        result = 'OBJECT'
    return result

def ftof(type):
    if type in ftof_dict: return ftof_dict[type]
    else: return 'TYPE(%s)'%type




#!/usr/bin/env python
# To use:
#       python setup.py install
#
import os, sys, stat
import subprocess

try:
    import distutils
    from distutils.core import setup
    from distutils.command.install import INSTALL_SCHEMES
except:
    raise SystemExit("Distutils problem")

try:
    from distutils.command.build_py import build_py_2to3 as build_py
except ImportError:
    from distutils.command.build_py import build_py

try:
    perm644 = stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH | stat.S_IWUSR
    os.chmod('source/License.txt',perm644)
    os.chmod('source/Forthon.h',perm644)
    os.chmod('source/Forthon.c',perm644)
except:
    print('Permissions on License.txt and Src files needs to be set by hand')

# --- Write out version information to the version.py file.
version = '0.8.25'
try:
    # --- In python3, check_output or Popen returns a byte string that needs to be decoded to get the string.
    # --- The decode method is mostly harmless in python2.
    #bcommithash = subprocess.check_output(['git', 'log', '-n', '1', '--pretty=%h'], stderr=subprocess.STDOUT).strip()
    # --- Needed for Py2.6 (which doesn't have subprocess.check_output)
    bcommithash = subprocess.Popen(['git', 'log', '-n', '1', '--pretty=%h'], stderr=subprocess.PIPE, stdout=subprocess.PIPE).communicate()[0].strip()
    commithash = bcommithash.decode()
    if not commithash:
        # --- If git is not available, commithash will be an empty string.
        raise OSError('git commit hash not found')
except (subprocess.CalledProcessError, OSError):
    # --- Error returned by subprocess.check_output if git command failed
    # --- This version was obtained from a non-git distrobution. Use the
    # --- saved commit hash from the release.
    # --- This is automatically updated by version.py.
    commithash = '35d45ac'

with open('source/version.py','w') as ff:
    ff.write("version = '%s'\n"%version)
    ff.write("gitversion = '%s'\n"%commithash)

# --- Get around a "bug" in disutils on 64 bit systems. When there is no
# --- extension to be installed, distutils will put the scripts in
# --- /usr/lib/... instead of /usr/lib64.
if distutils.sysconfig.get_config_vars()["LIBDEST"].find('lib64') != -1:
    for scheme in INSTALL_SCHEMES.values():
        scheme['purelib'] = scheme['platlib']

if sys.hexversion < 0x03000000:
    Forthonroot = 'Forthon'
elif sys.hexversion >= 0x03000000:
    Forthonroot = 'Forthon3'

# --- Normally, the package building script is called Forthon, but on Windows,
# --- it works better if it is called Forthon.py.
if sys.platform == 'win32':
    Forthon = Forthonroot+'.bat'
    ff = open(Forthon,'w')
    file = """\
@echo off
set sys_argv=
:Loop
if "%%1"=="" GOTO Continue
set sys_argv=%%sys_argv%% %%1
SHIFT
GOTO Loop
:Continue
%s -c "import Forthon.Forthon_builder" %%sys_argv%%
"""%(sys.executable)
    ff.write(file)
    ff.close()
else:
    Forthon = Forthonroot

# --- Force the deletion of the build directory so that a fresh install is
# --- done every time. This is needed since otherwise, after the first install,
# --- each subsequent install would use the same Forthon script and not update
# --- the python path in it appropriately.
os.system("rm -rf build")

setup (name = "Forthon",
       version = version,
       author = 'David P. Grote',
       author_email = "DPGrote@lbl.gov",
       license = "BSD-3-Clause-LLNL",
       url = "http://hifweb.lbl.gov/Forthon",
       description = "Fortran95 wrapper/code development package",
       long_description = """
Forthon provides an extensive wrapping of Fortran95 code, giving access to
routines and to any data in Fortran modules. Forthon also
provides an extensive wrapping of Fortran derived types, giving access to
derived type members, allowing passing of derived types into Fortran routines,
and creation of instances at the Python level. A mechanism for automatic
building of extension modules is also included. Versions using Numeric and
Numpy are available.""",
       platforms = 'any',
       classifiers = [
        'Programming Language :: Python',
        'Development Status :: 5 - Production/Stable',
        'Natural Language :: English',
        'Environment :: Console',
        'Intended Audience :: Science/Research',
        'License :: Free To Use But Restricted',
        'Operating System :: OS Independent',
        'Topic :: Software Development :: Build Tools',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        ],
       packages = ['Forthon'],
       package_dir = {'Forthon': 'source'},
       package_data = {'Forthon': ['License.txt','Forthon.h','Forthon.c']},
       scripts = [Forthon],
       cmdclass = {'build_py':build_py}
       )

# --- Clean up the extra file created on win32.
if sys.platform == 'win32':
    os.system("rm -f Forthon.py")




# Created by David P. Grote, March 6, 1998
# Modified by T. B. Yang, May 19, 1998
# Parse the interface description file

# This reads in the entire variable description file and extracts all of
# the variable and subroutine information needed to create an interface
# to python (or other scripting language).

import sys
import fvars
import re
if sys.version[0] == '1':
    import regsub

attribute_pat = re.compile('[ \t\n]*([a-zA-Z_]+)')

def processfile(packname, filename, othermacros=[], timeroutines=0):

    # Open the variable description file
    varfile = open(filename, 'r')

    # Read in the whole thing
    # (Make sure that there is a newline character at the end)
    text = varfile.read() + '\n'
    varfile.close()

    # Check if any other files where included in the list
    textother = []
    for a in othermacros:
        f = open(a, 'r')
        t = f.read()
        f.close()
        textother = textother + [t]

    # Get the package name (first line of file) and strip it from the text
    i = re.search('[\n \t#]', text).start()
    line = text[:i]
    text = text[i+1:].strip()

    # Check to make sure that the package name is correct
    if packname != line:
        print """

    Warning: the package name in the file does not agree
    with the file name. Make sure that the first line of the
    variable description file is the package name.
    file name = %r
    packagename = %r

    One possible reason this error is that the variable description
    file is in dos format. If this is so, change the format to match
    your system. (This can be done in vi by opening the file and
    typing ":set fileformat=unix" and then saving the file.)

    """%(packname, line)
        return []

    # Remove all line continuation marks
    text = re.sub('\\\\', '', text)

    # Get rid of any initial comments and blank lines
    while len(text) > 0 and text[0] ==  '#':
        text = text[re.search('\n', text).start()+1:].strip()

    # Get macros (everthing between curly braces) from other files and
    # prepend to current file (inside curly braces).
    for t in textother:
        m = t[re.search('\n{', t).start()+2:re.search('\n}', t).start()+1]
        if m:
            if text[0] == '{':
                text = '{' + m + text[1:]
            else:
                text = '{' + m + '}\n' + text

    # Deal with statements in between initial curly braces
    # (Only macro statements are used, everything else is ignored.)
    if len(text) > 0 and text[0] == '{':
        while text[0] != '}':
            if 'a' <= text[0] and text[0] <= 'z' or 'A' <= text[0] and text[0] <= 'Z':
                macro = text[0:re.search('=', text).start()].strip()
                text = text[re.search('=', text).start()+1:].strip()
                value = text[0:re.search('[#\n]', text).start()].strip()
                value = repr(eval(value))
                if sys.version[0] == '1':
                    text = regsub.gsub('\<'+macro+'\>', value, text)
                else:
                    text = re.sub('(?<=\W)'+macro+'(?=\W)', value, text)
                text = re.sub('/'+macro+'/', '/'+value+'/', text)
            text = text[re.search('\n', text).start()+1:].strip()
        text = text[re.search('\n', text).start()+1:].strip()

    # Get rid of any comments and blank lines
    while len(text) > 0 and text[0] ==  '#':
        text = text[re.search('\n', text).start()+1:].strip()

    # Parse rest of file, gathering variables

    # Create blank list to put the variables in.
    vlist = []

    # Create blank list to put the 'hidden' variables in.
    hidden_vlist = []

    # Create list of types
    typelist = []

    group = ''
    attributes = ''

    # The parser needs to distinguish between variable names and type names.
    # When readyfortype is 0, the next string found will be interpreted as a
    # variable name. If 1, then it will be a type name.
    readyfortype = 0

    # Parse rest of file
    hidden = 0
    istype = 0
    while text:
        text = text + '\n'

        # Check if group
        if text[0] == '*':
            istype = 0
            # Check the syntax
            g = re.match('\*+ (\w+)((?:[ \t]+\w+)*):', text)
            if g is None:
                g = re.match('(.*)', text)
                raise SyntaxError('Line defining the group is invalid\n%s'%g.group(1))
            # Then get new group name
            i = re.search(':', text).start()
            g = text[:i].split()
            group = g[1]
            # Include group name as an attribute
            attributes = ' '+' '.join(g[1:])+' '
            # Look for the 'hidden' attribute
            hidden = 0
            if re.search(' hidden ', attributes) != None:
                hidden = 1
            # Strip off group name and any comments
            text = text[i+1:].strip() + '\n'
            while text[0] ==  '#':
                text = text[re.search('\n', text).start()+1:].strip() + '\n'
            # Set i so that nothing is stripped off at end of 'if' block
            i = -1
            readyfortype = 0

        # Check if derived type
        elif text[0] == '%':
            istype = 1
            # Then get new type name
            i = re.search(':', text).start()
            tname = text[:i].split()[1]
            group = tname
            # Include group name as an attribute
            attributes = ' '+tname+' '
            # Create new instance of Ftype and append to the list
            ftype = fvars.Ftype(tname, attributes)
            typelist.append(ftype)
            # Strip off group name and any comments
            text = text[i+1:].strip() + '\n'
            while text[0] ==  '#':
                text = text[re.search('\n', text).start()+1:].strip() + '\n'
            # Set i so that nothing is stripped off at end of 'if' block
            i = -1
            readyfortype = 0

        # Check if variable is dynamic
        elif text[0] == '_':
            v.dynamic = 1
            i = 0

        # Check if type is real
        elif re.match('real\s', text):
            v.type = 'real'
            i = 3
            readyfortype = 0

        # Check if type is double
        elif re.match('double\s', text):
            v.type = 'double'
            i = 5
            readyfortype = 0

        # Check if type is float
        elif re.match('float\s', text):
            v.type = 'float'
            i = 4
            readyfortype = 0

        # Check if type is integer
        elif re.match('integer\s', text):
            v.type = 'integer'
            i = 6
            readyfortype = 0

        # Check if type is logical
        elif re.match('logical\s', text):
            v.type = 'logical'
            i = 6
            readyfortype = 0

        # Check if type is Filedes (temporary fix for now)
        elif re.match('Filedes\s', text):
            v.type = 'integer'
            i = 6
            readyfortype = 0

        # Check if type is Filename (assumed to a string of length 256)
        elif re.match('Filename\s', text):
            v.type = 'character'
            i = 7
            v.dims = ['256'] + v.dims
            readyfortype = 0

        # Check if type is character
        elif re.match('character[\s*]', text):
            v.type = 'character'
            v.array = 1
            i = re.search('[ \t\n]', text).start()
            if text[9] == '*':
                v.dims = [text[10:i]] + v.dims
            readyfortype = 0

        # Check if type is complex
        elif re.match('complex\s', text):
            v.type = 'complex'
            i = 6
            readyfortype = 0
           #if not v.dims:
           #  v.dims = ['1']

        # Check if variable is a function
        elif re.match('function\s', text):
            v.function = 'fsub'
            v.array = 0
            i = 7
            readyfortype = 0
            if timeroutines:
                # --- Add variable used to accumulate the time
                timerv = fvars.Fvars()
                timerv.name = v.name+'runtime'
                timerv.type = 'real'
                timerv.data = '/0./'
                timerv.unit = 'seconds'
                timerv.comment = 'Run time for function %s'%v.name
                timerv.group = group
                timerv.attr = attributes
                vlist.append(timerv)

        # Check if variable is a subroutine
        elif re.match('subroutine\s', text):
            v.function = 'fsub'
            v.array = 0
            v.type = 'void'
            i = 9
            readyfortype = 0
            if timeroutines:
                # --- Add variable used to accumulate the time
                timerv = fvars.Fvars()
                timerv.name = v.name+'runtime'
                timerv.type = 'real'
                timerv.data = '/0./'
                timerv.unit = 'seconds'
                timerv.comment = 'Run time for subroutine %s'%v.name
                timerv.group = group
                timerv.attr = attributes
                vlist.append(timerv)

        # Check if variable is a C subroutine (takes C ordered arrays)
        elif re.match('csubroutine\s', text):
            v.function = 'csub'
            v.array = 0
            v.type = 'void'
            i = 10
            readyfortype = 0
            if timeroutines:
                # --- Add variable used to accumulate the time
                timerv = fvars.Fvars()
                timerv.name = v.name+'runtime'
                timerv.type = 'real'
                timerv.data = '/0./'
                timerv.unit = 'seconds'
                timerv.comment = 'Run time for C subroutine %s'%v.name
                timerv.group = group
                timerv.attr = attributes
                vlist.append(timerv)

        # Check if variable is a parameter, i.e. not writable
        elif re.match('parameter\s', text):
            if v.array or v.function:
                raise SyntaxError('%s: only scalar variables can be a parameter'%v.name)
            v.parameter = 1
            i = 8

        # Check if there are any dimensions
        elif text[0] == '(':
            v.array = 1
            i = findmatchingparenthesis(0, text, v.name)
            v.dimstring = text[0:i+1]
            v.dims = v.dims + convertdimstringtodims(v.dimstring)

        # Look for a data field
        elif text[0] == '/':
            i = 1
            # If the data is a string, skip over the string in case the
            # '/' character appears, as in a date for example.
            if   text[i] == '"': i = re.search('"', text[2:]).start() + 2
            elif text[i] == "'": i = re.search("'", text[2:]).start() + 2
            i = re.search('/', text[i:]).start() + i
            data = text[0:i+1]
            # Handle the old Basis syntax for initial logical values.
            if data[1:-1] == 'FALSE': data = '/.false./'
            if data[1:-1] == 'TRUE': data = '/.true./'
            v.data = data
            readyfortype = 0

        # Look for a unit field
        elif text[0] == '[':
            i = re.search(']', text).start()
            v.unit = text[1:i]
            readyfortype = 0

        # Look for a limited field
        elif text[0:7] == 'limited':
            j = re.search('\(', text).start()
            i = findmatchingparenthesis(j, text, v.name)
            v.limit = text[j:i+1]
            readyfortype = 0

        # Look for attribute to add
        elif text[0] == '+':
            m = attribute_pat.match(text[1:])
            i = m.end() + 1
            if i != -1:
                v.attr = v.attr + m.group(1) + ' '
            readyfortype = 0

        # Look for attribute to subtract
        elif text[0] == '-':
            m = attribute_pat.match(text[1:])
            if m != None:
                i = m.end() + 1
                if sys.version[0] == '1':
                    v.attr = regsub.sub('\<'+m.group(1)+'\>', ' ', v.attr)
                else:
                    v.attr = re.sub('(?<=\W)'+m.group(1)+'(?=\W)', ' ', v.attr, count=1)
            else:
                i = 0
            readyfortype = 0

        # Look for a set action flag
        elif re.match('SET\s', text):
            v.setaction = 1
            i = 2
            readyfortype = 0

        # Look for a get action flag
        elif re.match('GET\s', text):
            v.getaction = 1
            i = 2
            readyfortype = 0

        # Look for comment (and remove extra spaces)
        elif text[0] == '#':
            i = re.search('\n', text).start() - 1
            v.comment = v.comment + text[1:i+2].lstrip()
            readyfortype = 0

        # Look for private remark
        elif text[0] == '$':
            i = re.search('\n', text).start() - 1
            readyfortype = 0

        # This only leaves a variable name or a new type
        else:
            if not readyfortype:
                # Create new variable
                v = fvars.Fvars()
                if not hidden and not istype:
                    vlist.append(v)
                elif hidden:
                    hidden_vlist.append(v)
                elif istype:
                    ftype.addvar(v)
                i = re.search('[ (\t\n]', text).start() - 1
                v.name = text[:i+1]
                v.group = group
                v.attr = attributes
                readyfortype = 1
            else:
                readyfortype = 0
                i = re.search('[ \t\n]', text).start() - 1
                v.type = text[:i+1]
                v.derivedtype = 1

        # Strip off field which was just parsed
        try:
            text = text[i+1:].strip()
        except IndexError:
            text = ''

    def processvar(v):
        # Use implicit typing if variable type was not set.
        if v.type == '':
            v.type = 'real'
            if 'i' <= v.name[0] and v.name[0] <= 'n':
                v.type = 'integer'

        # Parse the dimensions of arrays
        if v.array:
            assert v.dims, "%s dimensions not set properly"%v.name
            dims = v.dims
            v.dims = []
            for d in dims:
                sd = d.split(':')
                fd = fvars.Fdims()
                if len(sd) == 1:
                    fd.low = '1'
                    fd.high = sd[0]
                else:
                    fd.low = sd[0]
                    fd.high = sd[1]
                # Check for dimensions which are expressions
                if v.dynamic:
                    if fd.high == '':
                        # Flag the dynamic arrays which do not have dimensions
                        # specified. Give it the fassign attribute since it is
                        # assumed that a shapeless variable will be assigned
                        # to in fortran.
                        v.dynamic = 3
                        v.attr = v.attr + ' fassign '
                    elif (re.search(fd.low , '[/\*\+\-]') != None or
                          re.search(fd.high, '[/\*\+\-]') != None):
                        # Flag the dynamic arrays which have an expression for a dimension
                        v.dynamic = 2
                else:
                    # For static arrays, evaluate the dimensions or remove the
                    # expression
                    if re.search(fd.low , '[/\*\+\-]') != None:
                        fd.low = repr(eval(fd.low))
                    if re.search(fd.high, '[/\*\+\-]') != None:
                        fd.high = repr(eval(fd.high))
                v.dims = v.dims + [fd]
            #v.dims.reverse()

        # Set v.args if variable is a function
        if v.function:
            # Remove a possible blank first argument
            if len(v.dims) > 0 and v.dims[0]=='': del v.dims[0]
            # Extract type from arguments, this includes possible dimension
            # descriptions of arrays
            # --- dimvars holds the list of arguments that are used in array
            # --- dimensions
            v.dimvars = []
            for a in v.dims:
                fa = fvars.Fargs()
                v.args = v.args + [fa]
                # --- parse the pattern name(dim1:dim2,...):type
                # --- Only the name must be given, the other parts are optional
                ma = re.search("[:(]|\Z", a)
                fa.name = a[:ma.start()]
                if ma.group() == '(':
                    i = findmatchingparenthesis(ma.start(), a, v.name)
                    fa.dimstring = a[ma.start():i+1]
                    dimlist = fa.dimstring[1:-1].split(',')
                    fa.dims = processargdimvars(dimlist, v.dimvars)
                    a = a[i+1:]
                    ma = re.search("[:]|\Z", a)
                else:
                    fa.dimstring = ''
                    fa.dims = []
                # --- Set the argument type
                if ma.group() == ':':
                    fa.type = a[ma.start()+1:]
                else:
                    # --- Use implicit typing if the type was not specified
                    fa.type = 'real'
                    if 'i' <= fa.name[0] and fa.name[0] <= 'n':
                        fa.type = 'integer'
            # --- Change the list of dimvars from names of the arguments to
            # --- references to the Fargs instance and position number
            i = 0
            for arg in v.args:
                if arg.name in v.dimvars:
                    assert arg.type in ['integer'], "Error in %s:Variables used in array dimensions for argument list must be integer type"%v.name
                    v.dimvars[v.dimvars.index(arg.name)] = [arg, i]
                i += 1
            # --- Make sure that only names that are arguments are used
            # --- This checks for leftover names that wern't replaced by the
            # --- above loop.
            for var in v.dimvars:
                if type(var) == type(''):
                    raise SyntaxError("%s: Only subroutine arguments can be specified as dimensions. Bad argument name is %s"%(v.name, var))
            # --- Empty out dims which is no longer needed (so it won't cause
            # --- confusion later).
            v.dims = []

        # Clean up the comment, removing extra spaces and replace " with '
        if v.comment:
            v.comment = re.sub(' +', ' ', v.comment)
            v.comment = re.sub('"', "'", v.comment)
            v.comment = v.comment.strip()

    # Do further processing on dims or arguments list, and check variable type
    for v in vlist: processvar(v)
    for t in typelist:
        for v in t.vlist: processvar(v)

    # Return the list
    return (vlist, hidden_vlist, typelist)

def findmatchingparenthesis(i, text, errname):
    # --- Note that text[i] should be "(".
    p = 1
    while p > 0:
        i = i + 1
        try:
            if text[i] == '(':
                p = p + 1
            elif text[i] == ')':
                p = p - 1
        except IndexError:
            print 'Error in subscript of variable '+errname
    return i

def convertdimstringtodims(dimstring):
    # --- Remove the beginning and ending parenthesis
    d = dimstring[1:-1]
    # --- Strip out all white space
    d = re.sub('[ \t\n]', '', d)
    # --- Remove optional argument signifying if it is the first character
    if len(d) > 0 and d[0] == ';':
        d = d[1:]
    # --- Remove other optional arguments
    d = re.sub(';', ',', d)
    # --- Carefully search through dimstring, looking for comma separated
    # --- arguments and for parenthetical blocks (that may contain commas
    # --- as well as other text, numbers, math symbols, and parenthesis).
    dimlist = []
    while len(d) > 0:
        m = re.search("[,(]", d)
        if m is not None and m.group() == '(':
            i0 = findmatchingparenthesis(m.start(), d, d) + 1
            m = re.search("[,(]", d[i0:])
        else:
            i0 = 0
        if m is None:
            i = len(d)
        elif m.group() == ',':
            i = i0 + m.start()
        dimlist = dimlist + [d[:i]]
        d = d[i+1:]
    return dimlist

def processargdimvars(dims, dimvars):
    # --- For each dimensions, finds the low and high pieces and finds all
    # --- of the names which appear.
    dimlist = []
    for d in dims:
        fd = fvars.Fdims()
        dimlist = dimlist + [fd]
        # --- Find the low and high pieces, which are separated by a colon
        sd = d.split(':')
        if len(sd) == 1:
            fd.low = '1'
            fd.high = sd[0]
        else:
            fd.low = sd[0]
            fd.high = sd[1]
         # --- This is not really needed since it can be done in C too.
         #try:
         #  low = repr(eval(fd.low))
         #  fd.low = low
         #except:
         #  pass
         #try:
         #  high = repr(eval(fd.high))
         #  fd.high = high
         #except:
         #  pass
        # --- Add any variable names to the list of variables in dimensions
        for dim in [fd.low, fd.high]:
            sl = re.split('[ ()/\*\+\-]', dim)
            for ss in sl:
                if re.search('[a-zA-Z]', ss) != None:
                    if ss not in dimvars: dimvars.append(ss)
    return dimlist

